///////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                     //
//       This software and its documentation are copyright (2013) by the     //
//   Broad Institute.  All rights are reserved.  This software is supplied   //
//   without any warranty or guaranteed support whatsoever. The Broad        //
//   Institute is not responsible for its use, misuse, or functionality.     //
///////////////////////////////////////////////////////////////////////////////


// RefTrace.  Assess HyperBasevector assembly hb using reference sequence G, by
// aligning each reference contig G[g] to hb.  Each such alignment consists of
// aligning segments, punctuated by gaps.  A number of choices are involved in
// defining these alignments.  End gaps are treated differently: we simply tally
// the number of unaligned bases.  Some gaps are likely not true gaps, but rather
// artifacts of the alignment process.
//
// HOW IT WORKS:
//
// 1. We build a larger HyperBasevector hbp in which every edge lives in both
//    orientations.
// 2. Each edge in hbp is aligned to the genome, allowing for multiple placements
//    (and subject to several heuristics).
// 3. To each alignment we associate its start and stop points on the genome.
//    The stop point is defined by finding the point on the genome that corresponds
//    to the point on the edge that is K-1 bases from its right end.  (This is not
//    well-behaved if there is an indel at this point.)
// 4  Because of 'random' choices in alignment, there may be small inconsistencies
//    between these points arising from an edge entering and an edge exiting a given
//    vertex.  We 'twiddle' the start and stop points slightly to get them to agree
//    in some cases (so long as cycles are not introduced).
// 5. We form a graph in which the vertices are certain triples (g,p,v) where g is
//    a reference contig id, p is a position on it, and v is a vertex id in hbp.
//    Each alignment defines two vertices and an edge between them.
// 6. We add edges to the graph to represent gaps, whenever we observe sink-source
//    'gaps' of size between -10000 and 10000 bases.  In the case of negative gaps,
//    this introduces some inaccuracy because of overcounting, although we eliminate
//    some of this below.
// 7. Then, without recomputing sources and sinks, we add 'master' sources and sinks
//    for each chromosome, and connect the master sources to the old sources and the
//    old sinks to the master sinks.
// 8. Each edge is assigned a penalty, 1*e + 100*g + 0.1*d, where
//    - e = alignment error count for the edge;
//    - g = the number of gaps;
//    - d = the number of gap bases.
// 9. We find shortest paths through the graph.  By construction there is a single
//    shortest path for each chromosome, which is our final answer.  Note that there
//    may be errors in common for each pair of edges * ----e1----> * ----e2----> * 
//    since they overlap by K-1 bases, which would result in overcounting.  Negative
//    gaps as described above could have the same effect.  This introduces some 
//    inaccuracy into the finding of shortest paths.  However having found the 
//    shortest paths, we then compensate (although not fully) by finding the total
//    nonredundant error count for a given path, in an appropriate sense.

// ISSUES:
//
// 1. As described above, there are several inconsistencies in the calculation
//    arising from indels and incompatible alignments.
// 2. Circular chromosomes are incorrectly handled.

// MakeDepend: library OMP
// MakeDepend: cflags OMP_FLAGS

#include "Basevector.h"
#include "CoreTools.h"
#include "Equiv.h"
#include "ParallelVecUtilities.h"
#include "PrintAlignment.h"
#include "graph/Digraph.h"
#include "kmers/KmerRecord.h"
#include "math/Functions.h"
#include "math/HoInterval.h"
#include "pairwise_aligners/SmithWatAffine.h"
#include "pairwise_aligners/SmithWatBandedA.h"
#include "paths/HyperBasevector.h"
#include "paths/long/SupportedHyperBasevector.h"
#include "paths/RemodelGapTools.h"
#include "paths/long/AssessBestAlignCore.h"
#include "paths/long/LongProtoTools.h"
#include "paths/long/MakeKmerStuff.h"
#include "paths/long/RefTrace.h"
#include "polymorphism/Edit.h"
#include "reporting/PerfStat.h"
#include "paths/long/Variants.h"
#include "paths/long/ReadOriginTracker.h"

namespace { // open anonymous namespace

// Traverse the graph and search for evidence of an edge's placement. Return
// true if small number of placement violation is found.
bool SearchSupports(const HyperBasevector& shb, const map<int, vec<int>>& edge_locs,
        int node, int offset, int depth, int nbad_bases) 
{
    const int MaxBaseMisPlacement = 1000;
    const int MaxDepth = 10;
    if (depth > MaxDepth) return true;
    if (nbad_bases > MaxBaseMisPlacement) return false;
    for (size_t i = 0; i < shb.From(node).size(); i++) {
        int next_node = shb.From(node)[i];
        int next_edge = shb.EdgeObjectIndexByIndexFrom(node, i);
        int next_offset = offset + shb.EdgeLengthKmers(next_edge);
        const int DevOffset = 100;
        bool agree = false;
        map<int, vec<int>>::const_iterator itc =
            edge_locs.find(next_edge);
        if (itc != edge_locs.end()) {
            for (int x: itc->second) 
                if (abs(x - next_offset) < DevOffset) 
                    agree = true;
        }
        if (SearchSupports(shb, edge_locs, next_node, next_offset, depth+1, nbad_bases 
                    + (agree ? 0 : shb.EdgeLengthKmers(next_edge)))) 
            return true;
    }
    return false;
}

int Penalty( const int e, const int g, const int d )
{    return int( round( 100 * ( e + 100*g + d/100.0 ) ) );    }

int CorrelatePositionsAlways( const align& a, const int x1 )
{    int pos1 = a.pos1( ), pos2 = a.pos2( );
     if ( x1 < pos1 ) return 0; // off the end, shouldn't happen
     if ( x1 == pos1 ) return pos2;
     int nblocks = a.Nblocks( );
     const avector<int> &gaps = a.Gaps( ), &lengths = a.Lengths( );
     for ( int j = 0; j < nblocks; j++ )
     {    if ( gaps(j) > 0 ) pos2 += gaps(j);
          if ( gaps(j) < 0 )
          {    for ( int x = 0; x < -gaps(j); x++ )
                    if ( x1 == pos1++ ) return pos2;    } // really in gap
          if ( x1 < pos1 + lengths(j) ) return pos2 + x1 - pos1;
          else
          {    pos1 += lengths(j), pos2 += lengths(j);    }    }
     return a.Pos2( );    } // off the end, shouldn't happen

// Try to twiddle alignment ends.  Be careful not to create cycles.

void Twiddle( const HyperBasevector& hbp,
     vec< quad< triple<int,int,int>, triple<int,int,int>, 
          pair<int,int>, pair<int,int> > >& vedata,
     const vec<int>& xto_left, const vec<int>& xto_right, const int max_twiddle )
{
     vec< vec<int> > left( hbp.N( ) ), right( hbp.N( ) );
     for ( int i = 0; i < vedata.isize( ); i++ )
     {    int  e = vedata[i].third.first;
          left[ xto_left[e] ].push_back(i), right[ xto_right[e] ].push_back(i);    }
     for ( int v = 0; v < hbp.N( ); v++ )
     {    while(1)
          {    Bool changed = False;
               for ( int j1 = 0; j1 < right[v].isize( ); j1++ )
               for ( int j2 = 0; j2 < left[v].isize( ); j2++ )
               {    int i1 = right[v][j1], i2 = left[v][j2];
                    if ( vedata[i1].first.first != vedata[i2].first.first ) continue;
                    int& x1 = vedata[i1].second.second; 
                    int& x2 = vedata[i2].first.second;
                    if ( x1 == x2 || Abs( x1 - x2 ) > max_twiddle ) continue;

                    /*
                    int e1 = vedata[i1].third.first, e2 = vedata[i2].third.first;
                    cout << "\ninconsistent alignments of edges " << e1
                         << " and " << e2 << "?\n\n";
                    const align &a1 = aligns[i1], &a2 = aligns[i2];
                    int gg = vedata[i1].first.first;
                    PRINT2( a1.pos2( ), a1.Pos2( ) );
                    PrintVisualAlignment( True, cout, hbp.EdgeObject(e1),
                         G[gg], a1 );
                    cout << "\n";
                    PRINT2( a2.pos2( ), a2.Pos2( ) );
                    PrintVisualAlignment( True, cout, hbp.EdgeObject(e2),
                         G[gg], a2 );
                    */

                    int M = Max( x1, x2 );

                    // Avoid cycles.

                    if ( !( vedata[i1].first.second < M ) ) continue;
                    if ( !( M < vedata[i2].second.second ) ) continue;

                    x1 = x2 = M;
                    changed = True;    }
               if ( !changed ) break;    }    }    }

void FindShortestPath( const int this_source, const int this_sink, 
     const vecbasevector& G, const digraphE<int>& Z, const digraphE<int>& ZS, 
     const vec<int>& suc, const vec< triple<int,int,int> >& egd, 
     const vec< triple<int,int,int> >& verts,
     const vec< triple< int, int, pair<int,int> > >& edges,
     const vec<align>& aligns, const HyperBasevector& hbp, vec< vec<int> >& spaths,
     vec< triple<int,int,int> >& spaths_egd, vec< pair<int,int> >& spaths_gg_pen,
     const int verbosity, ostream& out )
{
     vec<int> spath;
     double sclock = WallClockTime( );
     int s1 = BinPosition( suc, this_source ), s2 = BinPosition( suc, this_sink );
     vec<int> pre;
     ZS.GetPredecessors1( s2, pre );
     digraphE<int> ZSP = ZS.Subgraph(pre);
     int p1 = BinPosition(pre, s1), p2 = BinPosition(pre, s2);
     ZSP.ShortestPath( p1, p2, spath );
     for ( int l = 0; l < spath.isize( ); l++ )
          spath[l] = suc[ pre[ spath[l] ] ];
     String stime = TimeSince(sclock);

     // Compute total errors.  Note that this will not always provide the exact
     // correct answer.  An effort is made to prevent overcounting, but this can
     // result in undercounting.

     int e = 0, g = 0, d = 0;
     vec< pair<int,edit0> > edits;
     for ( int i = 0; i < spath.isize( ) - 1; i++ )
     {    int v = spath[i], w = spath[i+1], p = 1000000000;

          // Find the best edge from v to w.

          int best_edge = -1;
          for ( int l = 0; l < Z.From(v).isize( ); l++ )
          {    if ( Z.From(v)[l] != w ) continue;
               int eid = Z.EdgeObjectIndexByIndexFrom( v, l );
               int el = egd[eid].first, gl = egd[eid].second, dl = egd[eid].third;
               int pl = Penalty( el, gl, dl );
               if ( pl < p )
               {    best_edge = l;
                    p = pl;    }    } 
          int eid = Z.EdgeObjectIndexByIndexFrom( v, best_edge );
          if ( eid < edges.isize( ) )
          {    int gg = verts[v].first;
               const align& a = aligns[eid];
               int edge_id = edges[eid].third.first;
               int p1 = a.pos1( ), p2 = a.pos2( );
               const basevector& E = hbp.EdgeObject(edge_id);

               // Traverse the alignment of the edge to locate its errors.

               Bool first = True;
               for ( int j = 0; j < a.Nblocks( ); j++ ) 
               {    if ( a.Gaps(j) > 0 )  
                    {    if ( !( first && edits.nonempty( )
                              && p2 <= edits.back( ).first ) )
                         {    edits.push( p2, edit0( DELETION, a.Gaps(j) ) );    }
                         if ( first && edits.nonempty( )
                              && p2 >= edits.back( ).first )
                         {    first = False;    }
                         p2 += a.Gaps(j);    }
                    if ( a.Gaps(j) < 0 ) 
                    {    if ( !( first && edits.nonempty( )
                              && p2 <= edits.back( ).first ) )
                         {    edits.push( p2, edit0( INSERTION, 
                                   basevector( E, p1, -a.Gaps(j) )
                                   .ToString( ) ) );    }
                         if ( first && edits.nonempty( )
                              && p2 >= edits.back( ).first )
                         {    first = False;    }
                         p1 -= a.Gaps(j);    }
                    for ( int x = 0; x < a.Lengths(j); x++ ) 
                    {    if ( E[p1] != G[gg][p2] )
                         {    if ( !( first && edits.nonempty( )
                                   && p2 <= edits.back( ).first ) )
                              {    edits.push( p2, edit0( SUBSTITUTION, 
                                        (char) as_base(E[p1]) ) );    }
                              if ( first && edits.nonempty( )
                                   && p2 >= edits.back( ).first )
                              {    first = False;    }    }
                         ++p1; ++p2;    }    }    }
          g += egd[eid].second; d += egd[eid].third;    }
     int gg = verts[ spath.front( ) ].first;
     int gstart = verts[ spath.front( ) ].second;
     int gstop = verts[ spath.back( ) ].second;
     int M = -1;
     for ( int i = 0; i < edits.isize( ); i++ )
     {    if ( edits[i].first < M ) continue;
          const edit0& x = edits[i].second;
          M = edits[i].first;
          if ( x.etype == SUBSTITUTION ) e++;
          else if ( x.etype == DELETION ) e += x.n;
          else e += x.seq.size( );    }
     int pen = Penalty( e, g, d );
     #pragma omp critical
     {    if ( verbosity >= 3 ) 
               PRINT8_TO( out, stime, gg, gstart, gstop, pen, e, g, d );
          spaths.push_back(spath);
          spaths_egd.push( e, g, d );
          spaths_gg_pen.push( gg, pen );    }    }

void AnnouncePaths( const vec< vec<int> >& spaths, const int K, 
     const digraphE<int>& Z, const vec< triple<int,int,int> >& egd, 
     const vec< triple<int,int,int> >& verts,
     const vec< triple< int, int, pair<int,int> > >& edges,
     const vec< triple<int,int,int> >& spaths_egd, 
     const vec< pair<int,Bool> >& to_hb, const int verbosity, ostream& out )
{
     if ( verbosity >= 1 )
     {    out << "\nThere are " << spaths.size( ) << " paths:\n";
          for ( int i = 0; i < spaths.isize( ); i++ )
          {    const vec<int>& p = spaths[i];
               int gg = verts[ p.front( ) ].first;
               out << "[" << i+1 << "] " << gg << "." << verts[ p.front( ) ].second
                    << " (" << verts[ p.front( ) ].third << ") --> " << gg << "." 
                    << verts[ p.back( ) ].second << " (" 
                    << verts[ p.back( ) ].third << ")" << " [errs=" 
                    << spaths_egd[i].first << ",gaps=" << spaths_egd[i].second 
                    << ",gap_bases=" << spaths_egd[i].third << "]" << endl;    }    }
     if ( verbosity >= 3 )
     {    out << "\nDetailed path info:\n";
          for ( int i = 0; i < spaths.isize( ); i++ )
          {    const vec<int>& p = spaths[i];
               int gg = verts[ p.front( ) ].first;
               out << "\n[" << i+1 << "] " << gg << "." 
                    << verts[ p.front( ) ].second << " (" 
                    << verts[ p.front( ) ].third << ") --> " << gg << "." 
                    << verts[ p.back( ) ].second << " (" 
                    << verts[ p.back( ) ].third << ")" << " [errs=" 
                    << spaths_egd[i].first << ",gaps=" << spaths_egd[i].second 
                    << ",gap_bases=" << spaths_egd[i].third << "]\n\n";
               for ( int j = 0; j < p.isize( ) - 1; j++ )
               {    int v = p[j], w = p[j+1];
                    out << "   " << gg << "." << verts[ p[j] ].second << " (" 
                         << verts[ p[j] ].third << ") --> " << gg << "." 
                         << verts[ p[j+1] ].second
                              + ( j == p.isize( ) - 2 ? (K-1) : 0 )
                         << " (" << verts[ p[j+1] ].third << ")";
                    int p = 1000000000, best_edge = -1;
                    for ( int l = 0; l < Z.From(v).isize( ); l++ )
                    {    if ( Z.From(v)[l] != w ) continue;
                         int eid = Z.EdgeObjectIndexByIndexFrom( v, l );
                         int el = egd[eid].first, gl = egd[eid].second; 
                         int dl = egd[eid].third;
                         int pl = Penalty( el, gl, dl );
                         if ( pl < p )
                         {    best_edge = l;
                              p = pl;    }    }
                    if ( best_edge < 0 )
                    {    out << " -- DISCONTINUITY\n";
                         continue;    }
                    int eid = Z.EdgeObjectIndexByIndexFrom( v, best_edge );
                    String hbp_edge_id = "*", hb_edge_id = "*";
                    if ( eid < edges.isize( ) )
                    {    int x = edges[eid].third.first;
                         hbp_edge_id = ToString(x);
                         hb_edge_id = ToString( to_hb[x].first )
                              + ( to_hb[x].second ? "fw" : "rc" );    }
                    out << " [hbp_edge=" << hbp_edge_id << ",hb_edge=" 
                         << hb_edge_id << ",e=" << egd[eid].first << ",g=" 
                         << egd[eid].second << ",d=" << egd[eid].third << "]" 
                         << endl;    }    }    }    }

// Align edges of hbp to reference.

template<int L> void AlignEdgesToRef( 

     // input data for reference:
          const vecbasevector& G,
     // input data for assembly:
          const HyperBasevector& hbp, const vec< pair<int,Bool> >& to_hb, 
     // outputs:
          vec< quad< triple<int,int,int>, triple<int,int,int>, 
               pair<int,int>, pair<int,int> > >& vedata,
          vec<align>& aligns, vec<int>& xto_left, vec<int>& xto_right,
     // heuristics:
          const double min_cov_frac, const double max_error_rate,
          const int max_offset_diff, const double min_group_frac,
          const int offset_add, const int min_group_save, const Bool fix_bug,
     // logging:
          const long_logging& logc, const int verbosity, ostream& out )
{
     // Setup for alignment.

     vecbasevector all(G);
     vec< triple<kmer<L>,int,int> > kmers_plus;
     MakeKmerLookup0( all, kmers_plus );
     vec< kmer<L> > kmers( kmers_plus.size( ) );
     for ( int64_t i = 0; i < kmers_plus.jsize( ); i++ )
          kmers[i] = kmers_plus[i].first;
     hbp.ToLeft(xto_left), hbp.ToRight(xto_right);

     // Go through the edges of the (doubled) assembly.

     #pragma omp parallel for schedule(dynamic,1)
     for ( int i = 0; i < hbp.EdgeObjectCount( ); i++ )
     {    const basevector& e = hbp.EdgeObject(i);

          // For each kmer in the edge, find its hits to the reference and find
          // the kmers having the most hits.

          int nkmers = e.isize( ) - L + 1;
          vec< triple<int64_t,int64_t,int64_t> > locs(nkmers);
          vec<int> pos( nkmers, vec<int>::IDENTITY );
          kmer<L> x;
          for ( int j = 0; j < nkmers; j++ )
          {    x.SetToSubOf( e, j );
               int64_t low = LowerBound(kmers, x), high = UpperBound(kmers, x);
               locs[j].first = high - low;
               locs[j].second = low, locs[j].third = high;    }
          if (fix_bug) ReverseSortSync( locs, pos );
          else SortSync( locs, pos );

          // Determine cutoff 'top'.

          double mcf = min_cov_frac;
          if ( logc.REFTRACE_VARIANTS ) mcf = 0.6;
          int t = int( floor( nkmers * mcf ) ), top;
          for ( top = t + 1; top < nkmers; top++ )
               if ( locs[top].first > locs[t].first ) break;

          // Find the associated offsets.

          vec< pair<int,int> > offset;
          for ( int j = 0; j < top; j++ )
          {    for ( int64_t m = locs[j].second; m < locs[j].third; m++ )
               {    int g = kmers_plus[m].second, o = kmers_plus[m].third - pos[j];
                    offset.push( g, o );    }    }
          Sort(offset);

          // Form offsets into groups.

          vec< triple< int, int, pair<int,int>  > > og;
          int mod = max_offset_diff;
          if ( logc.REFTRACE_VARIANTS ) mod = 500;
          for ( int j = 0; j < offset.isize( ); j++ )
          {    int k;
               for ( k = j + 1; k < offset.isize( ); k++ )
               {    if ( offset[k].first != offset[j].first ) break;
                    if ( offset[k].second - offset[k-1].second > mod )
                         break;    }
               og.push( k - j, offset[j].first, 
                    make_pair( offset[j].second, offset[k-1].second ) );
               j = k - 1;    }
          ReverseSort(og);
          if ( verbosity >= 4 )
          {    
               #pragma omp critical
               {    out << "\noriginal edge " << to_hb[i].first << ": ";
                    PRINT4_TO( out, nkmers, top, 
                         offset.size( ), og.size( ) );    
                    for ( int j = 0; j < og.isize( ); j++ )
                         PRINT2_TO( out, j, og[j].first );    }    }

          // Filter offset groups.

          double mgf = min_group_frac;
          if ( logc.REFTRACE_VARIANTS ) mgf = 0.65;
          int gj;
          for ( gj = 0; gj < og.isize( ); gj++ )
          {    if ( og[gj].first < min_group_save
                    && og[gj].first < mgf * og[0].first ) 
               {    break;    }    }
          og.resize(gj);
          if ( verbosity >= 3 && og.nonempty( ) )
          {    
               #pragma omp critical
               {    out << "\noffsets for edge " << i << " (hb_edge=" 
                         << to_hb[i].first << ", nkmers=" << nkmers << ")" << endl;
                    for ( int j = 0; j < og.isize( ); j++ )
                    {    out << "[" << j << "] " << og[j].second << "."
                              << og[j].third.first << "-" << og[j].third.second 
                              << " (" << og[j].first << ")" << endl;    }    }    }

          // Align.  The reason for adding to the offset is that there could be in
          // indel in the first or last L bases.

          for ( int j = 0; j < og.isize( ); j++ )
          {    int g = og[j].second;
               int off_low = og[j].third.first, off_high = og[j].third.second;
               int mid_offset = ( off_low + off_high ) / 2;
               int bandwidth 
                    = Max(mid_offset - off_low, off_high - mid_offset) + offset_add;

               // Do the alignment.  This is kludgy.  If the alignment has too 
               // many errors and the edge is long, we suspect that the problem
               // might be with a big indel, so we align using a larger bandwidth.
               // Note the unfortunate us of hardcoded constants.

               align a;
               int errors;

               if ( !logc.REFTRACE_VARIANTS )
               {
                   const int SMA_method = 1;
                   if (SMA_method == 1) {
                       SmithWatBandedA( hbp.EdgeObject(i), G[g],
                            -mid_offset, bandwidth, a, errors, 0, 1, 1 );
                   } else if(SMA_method == 2) {
                       SmithWatAffineBanded( hbp.EdgeObject(i), G[g],
                               -mid_offset, bandwidth, a, errors );
                   } else {
                       cout << "unrecognized SMA_method" << endl;
                   }
                   if ( double(errors) / double( a.extent2( ) ) > max_error_rate )
                   {    const int long_edge = 5000;
                        const int max_indel = 5000;
                        if ( hbp.EdgeLengthBases(i) < long_edge ) continue;
                        SmithWatBandedA( hbp.EdgeObject(i), G[g],
                             -mid_offset, max_indel, a, errors, 0, 1, 1 );
                        if ( double(errors) / double( a.extent2( ) ) > max_error_rate )
                             continue;    }
               }
               else
               {    double score = SmithWatAffineBanded( hbp.EdgeObject(i), G[g],
                           -mid_offset, bandwidth, a, errors ) / 3.0;
                    if ( verbosity >= 3 )
                    {    
                         #pragma omp critical
                         {    double err_rate = score / double( a.extent2( ) );
                              int hb_edge = to_hb[i].first;
                              int offset = -mid_offset;
                              PRINT5( hb_edge, offset, bandwidth, score, err_rate );
                                   }    }
                    double var_max_error_rate = 0.3;
                    if ( score / double( a.extent2( ) ) > var_max_error_rate )
                         continue;    }

               if ( verbosity >= 3 )
               {
                    #pragma omp critical
                    {    out << "\nalignment " << j << " of edge " << i << " ("
                              << xto_left[i] << " --> " << xto_right[i] 
                              << ", hb_edge=" << to_hb[i].first << ")" << endl;
                         vec<int> errs 
                              = a.MutationsGap1Gap2( hbp.EdgeObject(i), G[g] );
                         int mismatches = errs[0];
                         int indels = errs[1] + errs[2];
                         PRINT5_TO( out, g, a.pos2( ), a.Pos2( ), 
                              mismatches, indels );
                         if ( verbosity == 4 )
                         {    PrintVisualAlignment( True, out, hbp.EdgeObject(i),
                                   G[g], a );    }
                         if ( verbosity >= 5 )
                         {    PrintVisualAlignment( False, out, hbp.EdgeObject(i),
                                   G[g], a );    }    }    }

               // Figure out where the position e.isize( ) - K + 1 should map to
               // under the alignment.  Note that because there could be an indel
               // there, this is not necessarily a meaningful answer.

               int x1 = e.isize( ) - hbp.K( ) + 1;
               int x2 = CorrelatePositionsAlways( a, x1 );

               // Save results.

               #pragma omp critical
               {    vedata.push( make_triple( g, a.pos2( ), xto_left[i] ),
                         make_triple( g, x2, xto_right[i] ),
                         make_pair( i, errors ), make_pair( a.pos1( ), a.Pos1( ) ) );
                    aligns.push_back(a);    }    }    }    
     // Sort the output to avoid the stochastic downstream behavior of BuildGraph
     // that seems depend on the input order of the alignment data.
     SortSync(vedata, aligns);
}

void AddGapEdges( digraphE<int>& Z, const vec< triple<int,int,int> >& verts,
     vec< triple<int,int,int> >& egd, const int min_dist,
     const int max_dist, const int verbosity, ostream& out
     ,const bool bPreserveDisconnectedComponents=false)
{
     vec<size_t> vertex_group(Z.N(),0);
     if(bPreserveDisconnectedComponents){
          vec<vec<int>> vertex_groups;
          Z.Components(vertex_groups);
          for(size_t gg=0;gg<vertex_groups.size();++gg){
              for( const auto& member: vertex_groups[gg]){
                  vertex_group[member]=gg;
              }
          }
     }
     vec<int> sources, sinks;
     Z.Sources(sources), Z.Sinks(sinks);
     if ( verbosity >= 3 )
     {    out << "\ninitial sources:\n";
          for ( int j = 0; j < sources.isize( ); j++ )
          {    int v = sources[j];
               out << "[" << j << "] " << verts[v].first << "."
                    << verts[v].second << " (" << verts[v].third << ")\n";    }
          out << "\ninitial sinks:\n";
          for ( int j = 0; j < sinks.isize( ); j++ )
          {    int v = sinks[j];
               out << "[" << j << "] " << verts[v].first << "."
                    << verts[v].second << " (" << verts[v].third << ")\n";    }    }
     vec< triple<int,int,int> > ss_data;
     for ( int j1 = 0; j1 < sources.isize( ); j1++ )
     {    int v = sources[j1];
          ss_data.push( verts[v].first, j1, 1 );    }
     for ( int j2 = 0; j2 < sinks.isize( ); j2++ )
     {    int w = sinks[j2];
          ss_data.push( verts[w].first, j2, 2 );    }
     Sort(ss_data);
     for ( int u = 0; u < ss_data.isize( ); u++ )
     {    int v;
          for ( v = u + 1; v < ss_data.isize( ); v++ )
               if ( ss_data[v].first != ss_data[u].first ) break;
          for ( int x1 = u; x1 < v; x1++ )
          {    if ( ss_data[x1].third != 1 ) continue;
               int j1 = ss_data[x1].second;
               for ( int x2 = u; x2 < v; x2++ )
               {    if ( ss_data[x2].third != 2 ) continue;
                    int j2 = ss_data[x2].second;
                    int v = sinks[j2], w = sources[j1];
                    if ( v == w ) continue;
                    int dist = verts[w].second - verts[v].second;
                    if ( dist < min_dist || dist > max_dist ) continue;
                    if ( dist < 0 ) dist = 0;
                    if (!bPreserveDisconnectedComponents || vertex_group[v]==vertex_group[w]){
                        Z.AddEdge( v, w, Penalty( 0, 1, dist ) );
                        egd.push( 0, 1, dist );
                    }
               }    }
          u = v - 1;    }    }

void AddSpecialVerts( const int K, const vec<int>& sources, const vec<int>& sinks,
     vec< triple<int,int,int> >& egd, const vecbasevector& G, digraphE<int>& Z,
     vec< triple<int,int,int> >& verts
     ,const bool bPreserveDisconnectedComponents=false)
{
     vec<size_t> vertex_group(Z.N(),0);
     vec<vec<int>> vertex_groups;
     if(bPreserveDisconnectedComponents){
          Z.Components(vertex_groups);
          for(size_t gg=0;gg<vertex_groups.size();++gg){
              for( const auto& member: vertex_groups[gg]){
                  vertex_group[member]=gg;
              }
          }
     }
     else{
         vertex_groups.push_back( vec<int>(Z.N()));
         for(size_t ii=0;ii<vertex_groups.back().size();++ii){ vertex_groups.back()[ii]=ii; }
     }
     int nz = Z.N( );
     const size_t nGroups = vertex_groups.size();
     ForceAssert(nGroups>0);

     Z.AddVertices( 2 * G.size( ) * nGroups);
     for( size_t group=0;group<nGroups;++group){
         for ( int g = 0; g < (int) G.size( ); g++ ){
              vertex_group.push_back(group);
              vertex_groups[group].push_back(verts.size());
              verts.push( g, 0, -1 );
         }
         for ( int g = 0; g < (int) G.size( ); g++ ){
              vertex_group.push_back(group);
              vertex_groups[group].push_back(verts.size());
              verts.push( g, G[g].size( ), -1 );
         }
     }
     for ( int j1 = 0; j1 < sources.isize( ); j1++ )
     {    int gg = verts[ sources[j1] ].first;
          int d = verts[ sources[j1] ].second;
          int g = ( d > 0 ? 1 : 0 );
          int group = vertex_group[sources[j1]];
//std::cout << nz + group*2*G.size() + gg << " " << sources[j1] << " " << Penalty( 0, g, d ) << std::endl;
          if(bPreserveDisconnectedComponents){
              Z.AddEdge( nz + group*2*G.size() + gg, sources[j1], Penalty( d, g, 0 ) );
              egd.push( d, g, 0 );
          }
          else{
              Z.AddEdge( nz + group*2*G.size() + gg, sources[j1], Penalty( 0, g, d ) );
              egd.push( 0, g, d );
          }
     }
     for ( int j2 = 0; j2 < sinks.isize( ); j2++ )
     {    int gg = verts[ sinks[j2] ].first;
          int d = Max( 0, G[gg].isize( ) - verts[ sinks[j2] ].second - (K-1) );
          int g = ( d > 0 ? 1 : 0 );
          int group = vertex_group[sinks[j2]];
//std::cout << sinks[j2] << " " << nz + group*2*G.size()+(int) G.size( ) + gg << Penalty( 0, g, d ) << std::endl;
          if(bPreserveDisconnectedComponents){
              Z.AddEdge( sinks[j2], nz + group*2*G.size()+(int) G.size( ) + gg, Penalty( d, g, 0 ) );
              egd.push( d, g, 0 );
          }
          else{
              Z.AddEdge( sinks[j2], nz + group*2*G.size()+(int) G.size( ) + gg, Penalty( 0, g, d ) );
              egd.push( 0, g, d );
          }
      }    }

void VariantPath( const vec< quad< triple<int,int,int>
                                 , triple<int,int,int>
                                 , pair<int,int>
                                 , pair<int,int>
                                 > >& vedata
                , const vec<align>& aligns
                , vec< triple<int,int,int> > verts
                , vec< triple< int, int, pair<int,int> > > edges
                , vec< triple<int,int,int> > egd
                , digraphE<int> Z
                , const int verbosity
                , ostream& out
                , ostream& callsOut
                , const vec< pair<int,Bool> >& hbp_to_hb
                , const int K
                , const vecbasevector& G
                , const vecbasevector& Gplus
                , const vec<int>& Gplus_ext
                , const HyperBasevector& hbp
                , const vec< pair<int,Bool> >& to_hb
                , const String& sVarFile
                , const String& sRefFile
                , const vec<String>& ref_tags
                , const vec<fastavector>& ref_seqs
                , const vec<size_t>& ref_shift
                , const ReadOriginTracker* p_read_tracker
                , const long_logging& logc
                )
{
     if ( verbosity ) cout << "In VariantPath" << std::endl;
     const int min_dist = -10000;
     const int max_dist = 10000;
/*
{
    ofstream zout("z_org.dot");
    vec<String> vl(Z.N());
    vec<vec<String>> el(Z.N());
    for( size_t ii=0;ii<Z.N();++ii){
        vl[ii]=ToString(ii);
        el[ii].resize(Z.FromSize(ii));
        for(size_t ee=0; ee< Z.FromSize(ii);++ee){
            el[ii][ee]=ToString(Z.EdgeObjectByIndexFrom(ii,ee));
        }
    }
    Z.DOT_vl(zout,vl,"",vec<vec<String>>(), vec<String>( ),el);
    zout.close();
}
*/
     vec<int> sources, sinks;
     Z.Sources(sources), Z.Sinks(sinks);
     AddGapEdges( Z, verts, egd, min_dist, max_dist, verbosity, out,true );
/*
{
    ofstream zout("z_ge.dot");
    vec<String> vl(Z.N());
    vec<vec<String>> el(Z.N());
    for( size_t ii=0;ii<Z.N();++ii){
        vl[ii]=ToString(ii);
        el[ii].resize(Z.FromSize(ii));
        for(size_t ee=0; ee< Z.FromSize(ii);++ee){
            el[ii][ee]=ToString(Z.EdgeObjectByIndexFrom(ii,ee));
        }
    }
    Z.DOT_vl(zout,vl,"",vec<vec<String>>(), vec<String>( ),el);
    zout.close();
}
*/
     AddSpecialVerts( K, sources, sinks, egd, G, Z, verts,true );
/*
{
    ofstream zout("z_sv.dot");
    vec<String> vl(Z.N());
    vec<vec<String>> el(Z.N());
    for( size_t ii=0;ii<Z.N();++ii){
        vl[ii]=ToString(ii);
        el[ii].resize(Z.FromSize(ii));
        for(size_t ee=0; ee< Z.FromSize(ii);++ee){
            el[ii][ee]=ToString(Z.EdgeObjectByIndexFrom(ii,ee));
        }
    }
    Z.DOT_vl(zout,vl,"",vec<vec<String>>(), vec<String>( ),el);
    zout.close();
}
*/
     if ( verbosity >= 3 ) cout << "\nPATHS:\n";
     Z.Sources(sources), Z.Sinks(sinks);
     vec< vec<int> > spaths;
     vec< triple<int,int,int> > spaths_egd;
     vec< pair<int,int> > spaths_gg_pen;
     #pragma omp parallel for
     for ( int j1 = 0; j1 < sources.isize( ); j1++ )
     {    vec<int> suc;
          Z.GetSuccessors1( sources[j1], suc );
          vec<int> sinksx = Intersection( sinks, suc );
          if ( verbosity >= 3 )
          {
               #pragma omp critical
               {    out << Date( ) << ": start source " << j1 << " of "
                         << sources.size( ) << ", successors = " << suc.size( )
                         << " of " << Z.N( ) << endl;    }    }
          digraphE<int> ZS = Z.Subgraph(suc);
          for ( int j2 = 0; j2 < sinksx.isize( ); j2++ )
          {    if ( sources[j1] == sinksx[j2] ) continue;
              // yin_debug

              if ( verbosity >= 2 )
              {
                  #pragma omp critical
                  {
              cout << "find shortest path between "
                  << sources[j1] << " edge= " << verts[sources[j1]].third
                  << " and " << sinksx[j2] << " edge= " << verts[sinksx[j2]].third
                  << endl;
                  }
              }
               FindShortestPath( sources[j1], sinksx[j2], G, Z, ZS, suc, egd,
                    verts, edges, aligns, hbp, spaths, spaths_egd,
                    spaths_gg_pen, verbosity, out );    }     }

     // Announce paths.
     SortSync( spaths_gg_pen, spaths, spaths_egd );
     AnnouncePaths( spaths, K, Z, egd, verts, edges, spaths_egd, to_hb, verbosity, out );

     enum {L_GID,L_FRONT,L_BACK,L_INDEX};
     typedef std::tuple<int64_t,int64_t,int64_t,size_t> loc_sort_t;
     vec<loc_sort_t> to_sort(spaths.size());
     for( size_t pp=0;pp<to_sort.size();++pp){
         const auto& path = spaths[pp];
         int64_t gid=-1;
         int64_t front = -1;
         int64_t back = -1;
         for( const auto& entry: path){
             if(verts[entry].third>=0){
                 if(gid==-1) gid = verts[entry].first;
                 if(front==-1) front = verts[entry].second;
                 back = verts[entry].second;
             }
         }
         to_sort[pp]=std::make_tuple(gid,front,back,pp);
     }
     Sort(to_sort);
     if(verbosity>0){
         std::cout<<"Sorted components: (gid front back index)"<<std::endl;
         for(const auto& entry: to_sort){
             std::cout << std::get<L_GID>(entry) << " "
                       << std::get<L_FRONT>(entry) << " "
                       << std::get<L_BACK>(entry) << " "
                       << std::get<L_INDEX>(entry) << std::endl;
         }
     }
     vec< vec<int> > spaths_processed;
     vec< triple<int,int,int> > spaths_egd_processed;
     vec< pair<int,int> > spaths_gg_pen_processed;
     for( size_t org=0; org<to_sort.size();){
//     int64_t last_gid=-1,last_back=-1;
//     for( size_t org=0; org<to_sort.size();++org){
         const auto org_gid = std::get<L_GID>(to_sort[org]);
         auto front = std::get<L_FRONT>(to_sort[org]);
         auto back = std::get<L_BACK>(to_sort[org]);
         const auto org_index = std::get<L_INDEX>(to_sort[org]);

//         ForceAssert( spaths[org_index].size() > 2);

         spaths_processed.push_back(vec<int>());
         spaths_egd_processed.push_back( spaths_egd[org_index]);
         spaths_gg_pen_processed.push_back( spaths_gg_pen[org_index]);
//         if( org_gid!=last_gid || back > last_back){
//             spaths_processed.push_back(spaths[org_index]);
//             spaths_egd_processed.push_back( spaths_egd[org_index]);
//             spaths_gg_pen_processed.push_back( spaths_gg_pen[org_index]);
//        }
         auto& new_path=spaths_processed.back();
         for(size_t ee=0;ee+1<spaths[org_index].size();++ee){ new_path.push_back(spaths[org_index][ee]); }
         size_t next=org+1;
         size_t sequence_back = org;
         for(;next < to_sort.size() && std::get<L_GID>(to_sort[next])==org_gid;++next){
             const auto next_index = std::get<L_INDEX>(to_sort[next]);
             const auto next_back = std::get<L_BACK>(to_sort[next]);

             // these need refinements

             ForceAssert( spaths_gg_pen_processed.back().first == spaths_gg_pen[next_index].first);

             if(next_back > back){
                 spaths_egd_processed.back().first+=spaths_egd[next_index].first;
                 spaths_egd_processed.back().second+=spaths_egd[next_index].second;
                 spaths_egd_processed.back().third+=spaths_egd[next_index].third;
                 spaths_gg_pen_processed.back().second+=spaths_gg_pen_processed[next_index].second;
                 ForceAssert( spaths[next_index].size() > 2);
                 {
                     auto v = new_path.back();
                     auto w = spaths[next_index][1];
                     int dist = verts[w].second - verts[v].second;
                     Z.AddEdge(v,w,max(0,dist));
                 }
                 for(size_t ee=1;ee+1<spaths[next_index].size();++ee){ new_path.push_back(spaths[next_index][ee]); }
                 back = next_back;
                 sequence_back = next;
             }
         }
         new_path.push_back( spaths[std::get<L_INDEX>(to_sort[sequence_back])].back() );
         org = next;
     }
     if(verbosity>0){
         std::cout<<"processed paths:\n"<<std::endl;
         for(const auto& new_path: spaths_processed){
             for(auto& entry: new_path){ std::cout << entry << " "; }
             std::cout << std::endl;
         }
         {
             ofstream zout("z_processed.dot");
             vec<String> vl(Z.N());
             vec<vec<String>> el(Z.N());
             for( decltype(Z.N()) ii=0;ii<Z.N();++ii){
                 vl[ii]=ToString(ii);
                 el[ii].resize(Z.FromSize(ii));
                 for(decltype(Z.FromSize(ii)) ee=0; ee< Z.FromSize(ii);++ee){
                     el[ii][ee]=ToString(Z.EdgeObjectByIndexFrom(ii,ee));
                 }
            }
             Z.DOT_vl(zout,vl,"",vec<vec<String>>(), vec<String>( ),el);
             zout.close();
         }
     }
     AnnouncePaths( spaths_processed, K, Z, egd, verts, edges, spaths_egd_processed, to_hb, verbosity, out );
     vec<vec< vec<int> >> best_path_s( G.size( ) ), eids_s( G.size( ) );
//     int best_g = -1;
     {    //vec<int> best_errs( G.size( ), 1000000000 ), best_i( G.size( ), -1 );
          if ( spaths_processed.empty( ) )
          {    out << "no best paths in VariantPath." << endl;
               _exit(1);    }
//          for ( int i = 0; i < spaths_processed.isize( ); i++ )
//          {    const vec<int>& p = spaths_processed[i];
//               int g = verts[ p.front( ) ].first;
//               if ( spaths_egd_processed[i].first < best_errs[g] )
//               {    best_errs[g] = spaths_egd_processed[i].first;
//                    best_i[g] = i;
//                    /*best_g = g;*/    }    }
 //         for ( int g = 0; g < (int) G.size( ); g++ )
//          {    if ( best_errs[g] == Min(best_errs) )
//               {    /*best_g = g*/;
//                    break;    }    }

          for( const auto& p: spaths_processed){
               int g = verts[ p.front( ) ].first;
//          for ( int g = 0; g < (int) G.size( ); g++ )
//          {    const vec<int>& p = spaths_processed[ best_i[g] ];
               best_path_s[g].push_back(vec<int>());
               eids_s[g].push_back(vec<int>());
               for ( int j = 0; j < p.isize( ) - 1; j++ )
               {    int v = p[j], w = p[j+1];
                    int p = 1000000000, best_edge = -1;
                    for ( int l = 0; l < Z.From(v).isize( ); l++ )
                    {    if ( Z.From(v)[l] != w ) continue;
                         int eid = Z.EdgeObjectIndexByIndexFrom( v, l );
                         int el = egd[eid].first, gl = egd[eid].second;
                         int dl = egd[eid].third;
                         int pl = Penalty( el, gl, dl );
                         if ( pl < p )
                         {    best_edge = l;
                              p = pl;    }    }
                    int eid = Z.EdgeObjectIndexByIndexFrom( v, best_edge );
                    if ( eid < edges.isize( ) )
                    {    best_path_s[g].back().push_back( edges[eid].third.first );
                         eids_s[g].back().push_back(eid);    }
                    }
               }
          }
//     int left_trim = vedata[ eids[best_g].front( ) ].fourth.first;
//     int right_trim = hbp.EdgeObject( best_path[best_g].back( ) ).isize( )
//          - vedata[ eids[best_g].back( ) ].fourth.second;

//     basevector best_bpath = hbp.EdgeObject( best_path[best_g][0] );
//     for ( int i = 1; i < best_path[best_g].isize( ); i++ )
//     {    best_bpath = TrimCat( hbp.K( ),
//               best_bpath, hbp.EdgeObject( best_path[best_g][i] ) );    }
//     best_bpath = basevector(
//          best_bpath, left_trim, best_bpath.isize( ) - left_trim - right_trim );
//     basevector best_ref = G[best_g];
//     vecbasevector best;
//     best.push_back(best_bpath);
//     best.push_back(best_ref);

     vec<vec<vec<pair<int,int>>>> ref_pos_all_s(G.size());
     for (size_t g = 0; g < ref_pos_all_s.size(); g++){
         ref_pos_all_s[g].resize(best_path_s[g].size());
         for( size_t c = 0 ; c < ref_pos_all_s[g].size() ; ++c){
           ref_pos_all_s[g][c].resize(best_path_s[g][c].size());
         }
     }
     for (int g = 0; g < (int)G.size( ); g++) {
         for(size_t c=0;c<best_path_s[g].size();++c){
//         std::cout << g << " " << c << std::endl;
             for (int i = 0; i < (int)best_path_s[g][c].size(); i++) {
                 int eid = eids_s[g][c][i];
                 int start_on_ref = vedata[eid].first.second;
                 int start_on_edge = vedata[eid].fourth.first;
                 ref_pos_all_s[g][c][i].first = start_on_ref - start_on_edge;
                 ref_pos_all_s[g][c][i].second = ref_pos_all_s[g][c][i].first +
                     hbp.EdgeLength(best_path_s[g][c][i]);
//std::cout << "pos:" << ref_pos_all_s[g][c][i].first << " " << ref_pos_all_s[g][c][i].second << std::endl;
             }
         }
     }
     vec<int> Gplus_ext_zero(G.size(), 0);
     const vecbasevector& Gplus_sel = (Gplus.empty() ? G : Gplus);
     const vec<int>& Gplus_ext_sel = (Gplus.empty() ? Gplus_ext_zero : Gplus_ext);
/*
     vec<vec<vec<int>>> best_paths;
     for ( const auto& entry: best_path){
         best_paths.push_back( vec<vec<int>>(1,entry));
     }
     */

     ReftraceVaraints(out, callsOut, G, Gplus_sel, Gplus_ext_sel, hbp,
             hbp_to_hb, ref_pos_all_s, best_path_s, sVarFile, sRefFile,
             ref_tags, ref_seqs, ref_shift, p_read_tracker,&logc,true);
     if ( verbosity ) cout << "finished VariantPath" << std::endl;

}

void BuildGraph(
     vec< quad< triple<int,int,int>, triple<int,int,int>, 
          pair<int,int>, pair<int,int> > >& vedata,
     vec<align>& aligns, vec< triple<int,int,int> >& verts,
     vec< triple< int, int, pair<int,int> > >& edges,
     vec< triple<int,int,int> >& egd, digraphE<int>& Z, const int verbosity,
     ostream& out, const vec< pair<int,Bool> >& hbp_to_hb )
{
     SortSync( vedata, aligns );
     for ( int i = 0; i < vedata.isize( ); i++ )
     {    edges.push( verts.size( ), verts.size( ) + 1, vedata[i].third );
          verts.push_back( vedata[i].first, vedata[i].second );    }
     if ( verbosity >= 2 ) PRINT2_TO( out, verts.size( ), edges.size( ) );
     vec<int> vid( verts.size( ), vec<int>::IDENTITY ), vidi( verts.size( ) );
     SortSync( verts, vid );
     for ( int j = 0; j < verts.isize( ); j++ )
          vidi[ vid[j] ] = j;
     vec<int> v_to_new( verts.size( ), 0 );
     vec<Bool> v_to_delete( verts.size( ), False );
     int vcount = -1;
     for ( int j = 0; j < verts.isize( ); j++ )
     {    if ( j == 0 || verts[j] != verts[j-1] ) vcount++;
          else v_to_delete[j] = True;
          v_to_new[j] = vcount;    }
     EraseIf( verts, v_to_delete );
     for ( int j = 0; j < edges.isize( ); j++ )
     {    edges[j].first = v_to_new[ vidi[ edges[j].first ] ];
          edges[j].second = v_to_new[ vidi[ edges[j].second ] ];    }
     if ( verbosity >= 2 )
     {    out << "\nvertices:\n";
          for ( int j = 0; j < verts.isize( ); j++ )
          {    out << "[" << j << "] " << verts[j].third << " : " 
                    << verts[j].first << "." << verts[j].second << endl;    }
          out << "\nedges:\n";
          for ( int j = 0; j < edges.isize( ); j++ )
          {    int g = verts[ edges[j].first ].first;
               int start = verts[ edges[j].first ].second;
               int stop = verts[ edges[j].second ].second;
               out << "[" << j << "] " << edges[j].first << " --> "
                    << edges[j].second << ", ref: " << g << "." << start << "-" 
                    << stop << ", len: " << stop - start << ", errs: " 
                    << edges[j].third.second << ", assembly edge: " 
                    << hbp_to_hb[ edges[j].third.first ].first << endl;    }    }
     int NN = verts.size( );
     vec< vec<int> > from(NN), to(NN), from_edge_obj(NN), to_edge_obj(NN);
     vec<int> penalty;
     for ( int j = 0; j < edges.isize( ); j++ )
     {    int v = edges[j].first, w = edges[j].second, errs = edges[j].third.second;
          from[v].push_back(w), to[w].push_back(v);
          from_edge_obj[v].push_back( penalty.size( ) );
          to_edge_obj[w].push_back( penalty.size( ) );
          penalty.push_back( Penalty( errs, 0, 0 ) );
          egd.push( errs, 0, 0 );    }
     for ( int v = 0; v < NN; v++ )
     {    SortSync( from[v], from_edge_obj[v] );
          SortSync( to[v], to_edge_obj[v] );    }
     Z.Initialize( from, to, penalty, to_edge_obj, from_edge_obj );    }

} // close anonymous namespace

RefTraceResults RefTrace( const vec<HyperBasevector>& GH, const vec<bool>& is_circular, 
     const vecbasevector& Gplus, const vec<int>& Gplus_ext,
     const SupportedHyperBasevector& hb, const vec<int>& inv, 
     const int verbosity, const int NUM_THREADS, const long_logging& logc, 
     ostream& out, ostream& callsOut, RefTraceHeuristics rth,
     const String& BEST_GLOBAL_OUT, const Bool fix_bug,
     const String& sVarFile, const String& sRefFile, const vec<String>& ref_tags, 
     const vec<fastavector>& ref_seqs, const vec<size_t>& ref_shift,
     const ReadOriginTracker* p_read_tracker)
{    
     double rclock = WallClockTime( );
     int K = hb.K( );
     // hb.TestValid(logc);
     if ( verbosity >= 2 )
          out << "assembly has " << hb.EdgeObjectCount( ) << " edges" << endl;
     ForceAssertEq( hb.EdgeObjectCount( ), inv.isize( ) );

     // Test and expand reference.

     vecbasevector G;
     vec<int> G_source;
     for ( int g = 0; g < (int) GH.size( ); g++ )
     {    const HyperBasevector& x = GH[g];
          if ( !x.Acyclic( ) )
               FatalErr("\nReference sequence must be acyclic.\nAbort.");
          vec<int> sources, sinks;
          x.Sources(sources), x.Sinks(sinks);
          if ( !sources.solo( ) || !sinks.solo( ) )
          {    FatalErr("\nEach reference sequence must have a unique source "
                         "and a unique sink.\n"
                         "Abort.");    }
          vec< vec<int> > paths;
          x.EdgePaths( sources[0], sinks[0], paths );
          for ( int j = 0; j < paths.isize( ); j++ )
          {    basevector z = x.EdgeObject( paths[j][0] );
               for ( int i = 1; i < paths[j].isize( ); i++ )
                    z = TrimCat( x.K( ), z, x.EdgeObject( paths[j][i] ) );
               G.push_back(z);    
               G_source.push_back(g);    }    }

     // Create a HyperBasevector hbp that equals hb plus its reverse complement.
     // However only do this for components that need it.

     if ( verbosity >= 1 ) out << Date( ) << ": building hbp" << endl;
     vec< vec<int> > compx;
     hb.Components(compx);
     vec<HyperBasevector> HB;
     vec< pair<int,Bool> > to_hb;
     for ( int i = 0; i < compx.isize( ); i++ )
     {    HB.push( hb, compx[i] );
          for ( int j = 0; j < compx[i].isize( ); j++ )
          {    int v = compx[i][j];
               for ( int l = 0; l < hb.From(v).isize( ); l++ )
                    to_hb.push( hb.EdgeObjectIndexByIndexFrom(v, l), True );    }
          if ( HB.back( ).EdgeObjectCount( ) > 0 
               && inv[ to_hb.back( ).first ] < 0 )
          {    HyperBasevector rc( HB.back( ) );
               rc.Reverse( );
               HB.push_back(rc);    
               for ( int j = 0; j < compx[i].isize( ); j++ )
               {    int v = compx[i][j];
                    for ( int l = 0; l < hb.From(v).isize( ); l++ )
                    {    to_hb.push( hb.EdgeObjectIndexByIndexFrom(v, l), 
                              False );    }    }    }    }
     HyperBasevector hbp( K, HB );
     if ( verbosity >= 2 )
     {    out << "'doubled' assembly has " << hbp.EdgeObjectCount( ) << " edges"
               << endl;    }

     // Map edges of hbp back to hb.  An edge in hbp maps to (hb_edge, fw?).

     vec< pair<int,Bool> > hbp_to_hb( hbp.EdgeObjectCount( ) );
     for ( int v = 0; v < hbp.N( ); v++ )
     {    for ( int j = 0; j < hbp.From(v).isize( ); j++ )
          {    int e = hbp.EdgeObjectIndexByIndexFrom( v, j );
               int ez = to_hb[e].first;
               Bool fw = to_hb[e].second;
               hbp_to_hb[e] = make_pair( ez, fw );    }    }

     // Heuristics.

     const int L = 40;
     const double min_cov_frac = 0.5;
     const int min_dist = -10000;
     const int max_dist = 10000;

     // Some data structures.
     // The vectors vedata and edges are synchronous.
     // The structure vedata has the following structure:
     // { (genome_tig_id, start_pos_on_genome_tig, left_vertex_of_edge_in_hbp ),
     //   (genome_tig_id, stop_pos_on_genome_tig-K+1, right_vertex_of_edge_in_hbp ),
     //   (hbp_edge_id, error_count),
     //   (start_pos_on_hbp_edge, stop_pos_on_hbp_edge) }.

     vec< quad< triple<int,int,int>, triple<int,int,int>, 
          pair<int,int>, pair<int,int> > > vedata;
     vec< triple< int, int, pair<int,int> > > edges;

     // Align edges of hbp to reference.

     vec<align> aligns;
     vec<int> xto_left, xto_right;
     AlignEdgesToRef<L>( G, hbp, to_hb, vedata, aligns, xto_left, xto_right,
          min_cov_frac, rth.max_error_rate, rth.max_offset_diff, rth.min_group_frac, 
          rth.offset_add, rth.min_group_save, fix_bug, logc, verbosity, out );

     // Try to twiddle alignment ends.  Be careful not to create cycles.

     Twiddle( hbp, vedata, xto_left, xto_right, rth.max_twiddle );

     // Build graph.

     vec< triple<int,int,int> > verts, egd;
     digraphE<int> Z;
     BuildGraph( vedata, aligns, verts, edges, egd, Z, verbosity, out, hbp_to_hb );

     if (logc.REFTRACE_VARIANTS_DC) {
         VariantPath( vedata, aligns, verts, edges, egd, Z, verbosity, out, callsOut , hbp_to_hb, K, G, Gplus, Gplus_ext, hbp,to_hb
                    , sVarFile , sRefFile , ref_tags , ref_seqs , ref_shift , p_read_tracker,logc);
     }

     // Add gap edges.  Then add special vertices for the ends of the chromosomes, 
     // and add gap edges from and to them.  Note that entries in verts refer to 
     // vertex "-1" of hbp.  Note that we deliberately do not recompute sources and 
     // sinks before this step.  That's because the previous step can introduce 
     // cycles.

     vec<int> sources, sinks;
     Z.Sources(sources), Z.Sinks(sinks);
     AddGapEdges( Z, verts, egd, min_dist, max_dist, verbosity, out );
     AddSpecialVerts( K, sources, sinks, egd, G, Z, verts );

     // Find shortest paths through the graph and score them.

     if ( verbosity >= 3 ) cout << "\nPATHS:\n";
     Z.Sources(sources), Z.Sinks(sinks);
     vec< vec<int> > spaths;
     vec< triple<int,int,int> > spaths_egd;
     vec< pair<int,int> > spaths_gg_pen;
     #pragma omp parallel for
     for ( int j1 = 0; j1 < sources.isize( ); j1++ )
     {    vec<int> suc;
          Z.GetSuccessors1( sources[j1], suc );
          vec<int> sinksx = Intersection( sinks, suc );
          if ( verbosity >= 3 )
          {
               #pragma omp critical
               {    out << Date( ) << ": start source " << j1 << " of "
                         << sources.size( ) << ", successors = " << suc.size( )
                         << " of " << Z.N( ) << endl;    }    }
          digraphE<int> ZS = Z.Subgraph(suc);
          for ( int j2 = 0; j2 < sinksx.isize( ); j2++ )
          {    if ( sources[j1] == sinksx[j2] ) continue;
              // yin_debug
              
              if ( verbosity >= 2 )
              {
              cout << "find shortest path between " 
                  << sources[j1] << " edge= " << verts[sources[j1]].third  
                  << " and " << sinksx[j2] << " edge= " << verts[sinksx[j2]].third 
                  << endl;
              }
               FindShortestPath( sources[j1], sinksx[j2], G, Z, ZS, suc, egd, 
                    verts, edges, aligns, hbp, spaths, spaths_egd, 
                    spaths_gg_pen, verbosity, out );    }     }

     // Announce paths.

     SortSync( spaths_gg_pen, spaths, spaths_egd );
     AnnouncePaths( spaths, K, Z, egd, verts, edges, spaths_egd, to_hb, 
          verbosity, out );
     REPORT_TIME( rclock, "used tracing reference - main" );

     // Print best global alignment.

     int meta_events = 0;
     if ( BEST_GLOBAL_OUT != "" || logc.SHOW_REFTRACE_EVENTS || logc.REFTRACE_VARIANTS)
     {    double gclock = WallClockTime( );
          if ( spaths.empty( ) )
          {    out << "BEST_GLOBAL_OUT: there are no paths, giving up." << endl;
               _exit(1);    }
          vec< vec<int> > best_path( G.size( ) ), eids( G.size( ) );
          int best_g = -1;
          {    vec<int> best_errs( G.size( ), 1000000000 ), best_i( G.size( ), -1 );
               for ( int i = 0; i < spaths.isize( ); i++ )
               {    const vec<int>& p = spaths[i];
                    int g = verts[ p.front( ) ].first;
                    if ( spaths_egd[i].first < best_errs[g] )
                    {    best_errs[g] = spaths_egd[i].first;
                         best_i[g] = i;   
                         best_g = g;    }    }
               for ( int g = 0; g < (int) G.size( ); g++ )
               {    if ( best_errs[g] == Min(best_errs) )
                    {    best_g = g;
                         break;    }    }
               for ( int g = 0; g < (int) G.size( ); g++ )
               {    const vec<int>& p = spaths[ best_i[g] ];
                    for ( int j = 0; j < p.isize( ) - 1; j++ )
                    {    int v = p[j], w = p[j+1];
                         int p = 1000000000, best_edge = -1;
                         for ( int l = 0; l < Z.From(v).isize( ); l++ )
                         {    if ( Z.From(v)[l] != w ) continue;
                              int eid = Z.EdgeObjectIndexByIndexFrom( v, l );
                              int el = egd[eid].first, gl = egd[eid].second; 
                              int dl = egd[eid].third;
                              int pl = Penalty( el, gl, dl );
                              if ( pl < p )
                              {    best_edge = l;
                                   p = pl;    }    }
                         int eid = Z.EdgeObjectIndexByIndexFrom( v, best_edge );
                         if ( eid < edges.isize( ) )
                         {    best_path[g].push_back( edges[eid].third.first );
                              eids[g].push_back(eid);    }    }    }    }

          int left_trim = vedata[ eids[best_g].front( ) ].fourth.first;
          int right_trim = hbp.EdgeObject( best_path[best_g].back( ) ).isize( )
               - vedata[ eids[best_g].back( ) ].fourth.second;

          basevector best_bpath = hbp.EdgeObject( best_path[best_g][0] );
          for ( int i = 1; i < best_path[best_g].isize( ); i++ )
          {    best_bpath = TrimCat( hbp.K( ), 
                    best_bpath, hbp.EdgeObject( best_path[best_g][i] ) );    }
          best_bpath = basevector( 
               best_bpath, left_trim, best_bpath.isize( ) - left_trim - right_trim );
          basevector best_ref = G[best_g];
          vecbasevector best;
          best.push_back(best_bpath);
          best.push_back(best_ref);
          if ( BEST_GLOBAL_OUT != "" ) best.WriteAll(BEST_GLOBAL_OUT);

          // call variants guided by the best path alignment on reference

          if (logc.REFTRACE_VARIANTS&&!logc.REFTRACE_VARIANTS_DC) {
              vec<vec<pair<int,int>>> ref_pos_all(G.size()); 
              for (size_t g = 0; g < ref_pos_all.size(); g++) 
                  ref_pos_all[g].resize(best_path[g].size());
              for (int g = 0; g < (int)G.size( ); g++) {    
                  for (int i = 0; i < (int)best_path[g].size(); i++) {
                      int eid = eids[g][i];
                      int start_on_ref = vedata[eid].first.second;
                      int start_on_edge = vedata[eid].fourth.first;
                      ref_pos_all[g][i].first = start_on_ref - start_on_edge;
                      ref_pos_all[g][i].second = ref_pos_all[g][i].first + 
                          hbp.EdgeLength(best_path[g][i]);
                  }
              }
              vec<int> Gplus_ext_zero(G.size(), 0);
              const vecbasevector& Gplus_sel = (Gplus.empty() ? G : Gplus);
              const vec<int>& Gplus_ext_sel = (Gplus.empty() ? Gplus_ext_zero : Gplus_ext);
              vec<vec< vec<int> >> best_path_s( G.size( ) );
              vec<vec<vec<pair<int,int>>>> ref_pos_all_s(G.size());

              for(size_t gg=0;gg<G.size();++gg){
                  best_path_s[gg].push_back( best_path[gg]);
                  ref_pos_all_s[gg].push_back( ref_pos_all[gg]);

              }

              ReftraceVaraints(out, callsOut, G, Gplus_sel, Gplus_ext_sel, hbp,
                      hbp_to_hb, ref_pos_all_s, best_path_s, sVarFile, sRefFile,
                      ref_tags, ref_seqs, ref_shift, p_read_tracker, &logc,logc.REFTRACE_VARIANTS_LIMIT_K1_EFFORT);
          }

          // Show RefTrace events.  This is part of what is in
          // auxmain/AssessBestAlign.cc.

          if (logc.SHOW_REFTRACE_EVENTS)
          {    
               for ( int g = 0; g < (int) G.size( ); g++ )
               {    
                    vec<pair<int,int>> best_bpath_pos(best_path[g].size());
                    basevector best_bpath = hbp.EdgeObject( best_path[g][0] );
                    best_bpath_pos[0] = make_pair(0, best_bpath.size());
                    for ( int i = 1; i < best_path[g].isize( ); i++ )
                    {    best_bpath = TrimCat( hbp.K( ), 
                              best_bpath, hbp.EdgeObject( best_path[g][i] ) );    
                         best_bpath_pos[i] = make_pair(best_bpath_pos[i-1].second - hbp.K() + 1,
                                 best_bpath.size());
                    }
                    best_bpath = basevector( best_bpath, left_trim, 
                         best_bpath.isize( ) - left_trim - right_trim );

                    const basevector& b1 = best_bpath, &b2 = G[g];
                    alignment al;
                    SmithWatAffineParallel( b1, b2, al );
                    align a(al);
                    vec< pair<int,int> > P1, P2;
                    DecomposeAlign( a, b1, b2, P1, P2 );
                    for ( int z = 0; z < P1.isize( ); z++ )
                    {    basevector B1( b1, P1[z].first, 
                              P1[z].second - P1[z].first );
                         basevector B2( b2, P2[z].first, 
                              P2[z].second - P2[z].first );
                         align A = a.TrimmedTo1( 
                              P1[z].first, P1[z].second - P1[z].first );
                         vec<int> mgg = A.MutationsGap1Gap2( B1, b2 );
                         int start = P2[z].first, stop = P2[z].second;
                         if(logc.SHOW_REFTRACE_EVENTS){
                             out << "\nMETA-EVENT " << ++meta_events << " (at " << g
                                  << "." << start << "-" << stop << ", subs = "
                                  << mgg[0] << ", dels = "
                                  << mgg[1] << ", ins = " << mgg[2] << ")\n";
                             PrintVisualAlignmentClean(
                                  False, out, B1, b2, A );

                         }
                    }
               }
          }
          REPORT_TIME( gclock, "used tracing reference - global" ); 

     }
     // Report gaps.

     double tclock = WallClockTime( );
     if ( verbosity >= 1 )
     {    out << "\ngaps:\n";
          for ( int i = 0; i < spaths.isize( ); i++ )
          {    const vec<int>& p = spaths[i];
               for ( int i = 0; i < p.isize( ) - 1; i++ )
               {    int v = p[i], w = p[i+1];
                    int min_penalty = 1000000000, best_edge = -1;
                    for ( int l = 0; l < Z.From(v).isize( ); l++ )
                    {    if ( Z.From(v)[l] != w ) continue;
                         int p = Z.EdgeObjectByIndexFrom( v, l );
                         if ( p < min_penalty )
                         {    min_penalty = p;
                              best_edge = Z.EdgeObjectIndexByIndexFrom( 
                                   v, l );    }    }
                    if ( best_edge < 0 ) continue;
                    if ( egd[best_edge].second == 0 ) continue;
                    int start = verts[v].second, stop = verts[w].second;
                    out << verts[v].first << "." << start << "-" << stop
                         << " [len = " << stop-start << "]\n";    }    }    }

     // Extract metrics from paths.

     int total_errs = 0, total_gaps = 0, total_gap_bases = 0;
     const int infinity = 100000000;
     vec<int> e_best( GH.size( ), infinity );
     vec<int> g_best( GH.size( ), infinity );
     vec<int> d_best( GH.size( ), infinity );
     for ( int i = 0; i < spaths.isize( ); i++ )
     {    int m = spaths_gg_pen[i].first;
          int x = G_source[m];
          e_best[x] = Min( e_best[x], spaths_egd[i].first );
          g_best[x] = Min( g_best[x], spaths_egd[i].second );
          d_best[x] = Min( d_best[x], spaths_egd[i].third );    }
     for ( int x = 0; x < (int) GH.size( ); x++ )
     {    if ( e_best[x] < infinity )
          {    total_errs += e_best[x];
               total_gaps += g_best[x];
               total_gap_bases += d_best[x];    }
          else
          {    total_gaps++;
               for ( int i = 0; i < G_source.isize( ); i++ )
               {    if ( G_source[i] == x )
                    {    total_gap_bases += G[i].size( );
                         break;    }    }    }    }

     // Print summary statistics.

     int penalty = Penalty( total_errs, total_gaps, total_gap_bases );
     int64_t gsize = 0;
     for ( int g = 0; g < (int) G.size( ); g++ )
          gsize += G[g].size( );
     double errors_per_Mb = 1000000.0 * double(total_errs) / double(gsize);
     int free_gaps = 0;
     for ( int g = 0; g < (int) G.size( ); g++ )
          if ( !is_circular[g] ) free_gaps += 2;
     double gaps_per_Mb
          = double(total_gaps - free_gaps) / ( double(gsize)/1000000.0 );
     int gap_bases_per
          = int( round( double(total_gap_bases) / double( G.size( ) ) ) );

     if(logc.REFTRACE_VARIANTS_SUMMARY){
         out << penalty << " penalty (p = 1*e + 100*g + 0.1*d)" << endl;
         out << "-[" << setiosflags(ios::fixed) << setprecision(2) << errors_per_Mb
              << resetiosflags(ios::fixed) << "%%%]- errors"
              << " (e = errs = " << total_errs << ")" << endl;
         if (logc.SHOW_REFTRACE_EVENTS)
              out << "-[" << meta_events << "]- error meta-events" << endl;
         out << "-[" << setiosflags(ios::fixed) << setprecision(2) << gaps_per_Mb
              << resetiosflags(ios::fixed) << "]- excess gaps per Mb "
              << "(g = total gaps = " << total_gaps
              << ", \"excess\" = " << total_gaps - free_gaps << ")" << endl;
         out << "-[" << gap_bases_per << "]- gap bases per reference contig "
              << "(d = gap bases = " << total_gap_bases << ")\n";
     }

     if ( logc.PERF_STATS )
     {    PerfStat::log( ) << std::fixed << setprecision(0)
               << PerfStat( "gap_bases_per",
               "gap bases per reference contig", gap_bases_per );
          PerfStat::log( ) << std::fixed << setprecision(2)
               << PerfStat( "errors_per_Mb", "errors per Mb", errors_per_Mb );
          PerfStat::log( ) << std::fixed << setprecision(2)
               << PerfStat( "gaps_per_Mb", "excess gaps per Mb", gaps_per_Mb );    }
     REPORT_TIME( tclock, "used tracing reference - tail" );
     return RefTraceResults( penalty, total_gaps, meta_events );    }




void RefTraceControl::ReadySampleLookup(){ //ready the pair manager
    if (read_head != "" && reads.empty()) {
        reads.ReadAll(read_head + ".fastb");
        quals.ReadAll(read_head + ".qualb");
        pm.Read(read_head + ".pairs");
        pm.makeCache();
        pm.printLibraryStats(std::cout);
        bPairsManagerInitialized = true;
        ForceAssertEq(reads.size(), quals.size());
        ForceAssertEq(reads.size(), pm.nReads());
        cout << Date() << ": Loaded " << reads.size() << " reads from " << 
            read_head << endl;
    }
}

vec<String> RefTraceControl::getSampleList()const{ //get a list of library names
    if( bPairsManagerInitialized) return pm.getLibraryNames();
    else                          return vec<String>();
};

int RefTraceControl::getSampleID(size_t read_ID)const { //get sample index from a read id
    int out= -1;
    if(bPairsManagerInitialized) { 
        int pair_id = pm.getPairID(read_ID);
        if (pair_id != -1)
            out= pm.libraryID(pair_id); 
    }
    return out;
}

String RefTraceControl::getSampleName(size_t read_ID)const { //get sample name from a read id
    String ans = "UNSPECIFIED";
    if(bPairsManagerInitialized){ 
        int pair_id = pm.getPairID(read_ID);
        if (pair_id != -1)
            ans = pm.libraryName(pair_id);
    }
    return ans;
};

