///////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                     //
//       This software and its documentation are copyright (2013) by the     //
//   Broad Institute.  All rights are reserved.  This software is supplied   //
//   without any warranty or guaranteed support whatsoever. The Broad        //
//   Institute is not responsible for its use, misuse, or functionality.     //
///////////////////////////////////////////////////////////////////////////////

#ifndef REFTRACE_H
#define REFTRACE_H

#include "CoreTools.h"
#include "paths/HyperBasevector.h"
#include "paths/long/LongProtoTools.h"
#include "PairsManager.h"
#include "Basevector.h"
#include "Qualvector.h"

class RefTraceResults {

     public:

     RefTraceResults( ) : penalty(0), gaps(0), meta_events(0) { }
     RefTraceResults( const int penalty, const int gaps, const int meta_events ) 
          : penalty(penalty), gaps(gaps), meta_events(meta_events) { }

     int penalty;
     int gaps;
     int meta_events;

};

class RefTraceHeuristics {

     public:

     RefTraceHeuristics( ) : max_offset_diff(10), max_error_rate(0.05), 
          offset_add(1), max_twiddle(3), min_group_frac(0.75), 
          min_group_save(1000000000) { }

     RefTraceHeuristics( const int max_offset_diffx, const double max_error_ratex,
          const int offset_addx, const int max_twiddlex,
          const double min_group_fracx, const int min_group_savex )
          : max_offset_diff(max_offset_diffx), max_error_rate(max_error_ratex),
          offset_add(offset_addx), max_twiddle(max_twiddlex),
          min_group_frac(min_group_fracx), min_group_save(min_group_savex)
     {    RefTraceHeuristics rth;
          if ( max_offset_diff < 0 ) max_offset_diff = rth.max_offset_diff;
          if ( max_error_rate < 0 ) max_error_rate = rth.max_error_rate;
          if ( offset_add < 0 ) offset_add = rth.offset_add;
          if ( max_twiddle < 0 ) max_twiddle = rth.max_twiddle;
          if ( min_group_frac < 0 ) min_group_frac = rth.min_group_frac;     
          if ( min_group_save < 0 ) min_group_save = rth.min_group_save;    }

     int max_offset_diff;
     double max_error_rate;
     int offset_add;
     int max_twiddle;
     double min_group_frac;
     int min_group_save;

};

class RefTraceControl{
public:
    RefTraceControl():sVariantOutFile(""),read_head(""),bPairsManagerInitialized(false){};
    RefTraceControl(const String& sOutFile):sVariantOutFile(sOutFile),read_head(""),bPairsManagerInitialized(false){};
    RefTraceControl(const String& sOutFile,const String& read_head) :sVariantOutFile(sOutFile),read_head(read_head),bPairsManagerInitialized(false){};
    virtual ~RefTraceControl(){};

    void setReadHead(const String& in){
        if(bPairsManagerInitialized){
            pm.clear();
            bPairsManagerInitialized=false;
        }
        read_head=in;
    };

    const String& getVariantOutFile()const{return sVariantOutFile;};
    const vec<String>& getRefTags()const{ return ref_tags; }
    const vec<size_t>& getRefStart()const{return ref_start;}
    const vec<fastavector>& getRefSeqs()const{return ref_seqs;}
    vec<String>& getRefTags(){ return ref_tags; }
    vec<fastavector>& getRefSeqs(){return ref_seqs;}
    vec<size_t>& getRefStart(){return ref_start;}

    //this is NOT thread safe because PairsManager is not thread safe
    void ReadySampleLookup();                   //ready the pair manager
    // one can easily consolidate multiple library into a single sample, to be worked out later
    vec<String> getSampleList()const;           //get the full list of sample names
    int getSampleID(size_t read_ID)const ;      //get sample index from a read id
    String getSampleName(size_t read_ID)const ; //get sample name from a read id
    const vecbasevector& Reads() const { return reads; }
    const vecqualvector& Quals() const { return quals; }

private:
    String sVariantOutFile;

    // one entry for each region entry
    vec<String> ref_tags;          // each entry is the sequence tag in REFERENCE
    vec<fastavector> ref_seqs;     // each entry is the sequence (trimmed according to REGION)
    vec<size_t> ref_start;         // start of the trimmed region in reference sequence

    String read_head;
    bool bPairsManagerInitialized;
    PairsManager pm;
    vecbasevector reads;
    vecqualvector quals;
};


// BEST_GLOBAL_OUT: if provided, name of fastb file to write two basevectors to:
// first the best assembly sequence, then the best reference sequence.

class SupportedHyperBasevector; // forward declaration
class ReadOriginTracker;

RefTraceResults RefTrace( const vec<HyperBasevector>& GH, const vec<bool>& is_circular, 
     const vecbasevector& Gplus, const vec<int>& Gplus_ext,
     const SupportedHyperBasevector& hb, const vec<int>& inv, 
     const int verbosity, const int NUM_THREADS, const long_logging& logc,
     ostream& out, ostream& callsOut, RefTraceHeuristics rth = RefTraceHeuristics( ),
     const String& BEST_GLOBAL_OUT = "", const Bool fix_bug = False,
     const String& sVarFile="", const String& sRefFile="OUT_GENOME",
     const vec<String>& ref_tags=vec<String>(), const vec<fastavector>& ref_seqs=vec<fastavector>(),
     const vec<size_t>& ref_shift=vec<size_t>(),
     const ReadOriginTracker* p_read_tracker = NULL
     );

#endif
