///////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                     //
//       This software and its documentation are copyright (2013) by the     //
//   Broad Institute.  All rights are reserved.  This software is supplied   //
//   without any warranty or guaranteed support whatsoever. The Broad        //
//   Institute is not responsible for its use, misuse, or functionality.     //
///////////////////////////////////////////////////////////////////////////////

// MakeDepend: library OMP
// MakeDepend: cflags OMP_FLAGS
#include "paths/long/VariantCallTools.h"

#include "CoreTools.h"
#include "Map.h"
#include "paths/long/Logging.h"
#include "paths/long/SupportedHyperBasevector.h"
#include "pairwise_aligners/SmithWatAffine.h"
#include "PrintAlignment.h"
#include "util/TextTable.h"
#include "graph/Digraph.h"
#include "graph/DigraphTemplate.h"
#include "paths/long/ReadOriginTracker.h"
#include "paths/long/EvalByReads.h"
#include <queue>

namespace {

void ExtendedBasevector(const HyperBasevector& hb, int edge, int ext_max,
        basevector& edge_extended, int& len_extended_left) 
{
    vec<int> to_left; hb.ToLeft(to_left);
    vec<int> to_right; hb.ToRight(to_right);

    vec<int> edges = {edge};

    int ext_left = 0;
    int node_left = to_left[edge];
    while (hb.To(node_left).size() > 0 && ext_left < ext_max) {
        int e_new = hb.EdgeObjectIndexByIndexTo(node_left, hb.To(node_left).size() -1);
        edges.push_front(e_new);
        ext_left += hb.EdgeLengthKmers(e_new);
        node_left = hb.To(node_left).back();
    }

    int ext_right = 0;
    int node_right = to_right[edge];
    while (hb.From(node_right).size() > 0 && ext_right < ext_max) {
        int e_new = hb.EdgeObjectIndexByIndexFrom(node_right, hb.From(node_right).size()-1);
        edges.push_back(e_new);
        node_right = hb.From(node_right).back();
        ext_right += hb.EdgeLength(e_new);
    }
    basevector edges_to_bases = hb.EdgePathToBases(edges);
    int trim_left = max(0, ext_left - ext_max);
    int trim_right = max(0, ext_right - ext_max);
    len_extended_left = ext_left - trim_left;
    edge_extended = basevector(edges_to_bases.begin() + trim_left,
                          edges_to_bases.end() - trim_right);
}

int CorrelatePositionsRangeChecked( const align& a, const int x1 )
{    
     int pos1 = a.pos1( ), pos2 = a.pos2( );
     if ( x1 < pos1 || x1 > a.Pos1()) return -1; // off the end, shouldn't happen
     if ( x1 == pos1 ) return pos2;
     int nblocks = a.Nblocks( );
     const avector<int> &gaps = a.Gaps( ), &lengths = a.Lengths( );
     for ( int j = 0; j < nblocks; j++ )
     {    if ( gaps(j) > 0 ) pos2 += gaps(j);
          if ( gaps(j) < 0 )
          {    for ( int x = 0; x < -gaps(j); x++ )
                    if ( x1 == pos1++ ) return pos2;    } // really in gap
          if ( x1 < pos1 + lengths(j) ) return pos2 + x1 - pos1;
          else
          {    pos1 += lengths(j), pos2 += lengths(j);    }    }
     return -1;
} // off the end, shouldn't happen

template<typename GraphT>
void DumpGraph(const String filename, const GraphT& shb) {
    // dump the simplified graph
    vec<String> edge_names2(shb.EdgeObjectCount(),"");
    vec<double> lengths2( shb.EdgeObjectCount( ) );
    for (size_t i = 0; i < edge_names2.size(); ++i) {
        edge_names2[i] = ToString(i);
        lengths2[i] = shb.EdgeLengthKmers(i);
    }
    ofstream dout(filename);
    shb.PrettyDOT( dout, lengths2, HyperBasevector::edge_label_info(
                HyperBasevector::edge_label_info::DIRECT, &edge_names2 ) );
}

// Find edits from the alignments, return pos1, pos2, edits triple, where
// pos1 is the position in source sequence (assembly), pos2 is the position
// in target (reference), and edit is the mutation from refernce to
// assembly.
// Generate insertion edits of unaligned head/tail source sequence
// if head_ins/tail_ins are set. We don't want to call those variants near the
// begin or end of the chromosome or chromosome segments.
void GetEditsFromAlign(const basevector& s, const basevector& t, const align& a, 
        vec<triple<int,int,String>>* edits, vec<pair<String,String>>* change = NULL, 
        char prev_char_s = 'X', char prev_char_t = 'X', bool head_ins = false,
        bool tail_ins = true)
{
    // prefix the two sequences by a leading base
    String ss(1, prev_char_t);
    //String ss(1, prev_char_s);
    ss += s.ToString();
    String tt(1, prev_char_t);
    tt += t.ToString();
    int p1 = a.pos1();
    int p2 = a.pos2();
    if (p1 > 0 && head_ins) { // insertion of p1 seq at beginning
        String alt = ss.substr(0, p1+1);
        String ref = tt.substr(p2,1);
        if (change != NULL)
            change->push(ref, alt);
        edits->push(p1, 0, "Ins: " + ss.substr(1,p1));
    }
    for (int j = 0; j < a.Nblocks(); ++j){
        int gap = a.Gaps(j);
        int len = a.Lengths(j);
        if ( gap > 0 ) {
            String ref = tt.substr(p2,gap+1); // equiv to p2-1 in t
            String alt = ss.substr(p1, 1);    // equiv to p1-1 in s
            alt[0] = ref[0];
            if (change != NULL)
                change->push(ref, alt);
            edits->push(p1, p2, "Del: "+ ToString(basevector(t,p2,gap)));
            p2 += gap;
        }
        if ( gap < 0 ) {
            int ndel = -gap;
            String ref = tt.substr(p2,1);
            String alt = ss.substr(p1, ndel+1);
            alt[0] = ref[0];
            if (change != NULL)
                change->push(ref, alt);
            edits->push(p1, p2, "Ins: " + ToString(basevector(s, p1, ndel)));
            p1 += ndel;
        }
        for ( int x = 0; x < len; x++ ) {
            if ( s[p1] != t[p2] ) {
                String ref = tt.substr(p2+1,1);
                String alt = ss.substr(p1+1,1);
                if (change != NULL)
                    change->push(ref, alt);
                edits->push(p1, p2, "Sub: " + ToString(basevector(s,p1,1)));
            }
            ++p1, ++p2;
        }
    }
    // insertion at the end
    if (p1 != (int)s.size() && tail_ins) {
        int nins = s.size() - p1;
        String ref = tt.substr(p2,1);
        String alt = ss.substr(p1, nins+1);
        alt[0] = ref[0];
        if (change != NULL)
            change->push(ref, alt);
        edits->push((int)s.size(), p2, "Ins: " + ss.substr(p1+1,nins));
    }
}

// Performe gapless alignment of reads to hyperbasevector and report the
// preference of reads aligned to each edge as calculated by qualsum difference
// between best in-edge placement and best out-edge placement.
void FindReadHomesBest(const vecbasevector& bases, const vecqualvector& quals, 
        const HyperBasevector& bubble_graph,
        vec< vec< pair<int,int> > >* homes_index,
        vec<int> edges_to_show_supports,
        vec<tuple<int,int,read_place>>* edge_rid_place,
        int verbosity = 0,const bool bSafeFindPlaces=false)
{
    if (verbosity >= 1)
        cout << Date() << ": Finding placements of " << bases.size() << " reads " << endl;

    ForceAssertEq(bases.size(), quals.size());
    const int L = 12;
    const int infinity = 1000000000;

    vec<int> to_left; bubble_graph.ToLeft(to_left);
    vec<int> to_right; bubble_graph.ToRight(to_right);

    HyperBasevector hb_fw(bubble_graph), hb_rc(bubble_graph);
    hb_rc.Reverse( );
    vec<int> to_right_fw, to_right_rc;
    hb_fw.ToRight(to_right_fw), hb_rc.ToRight(to_right_rc);
    vecbasevector x_fw, x_rc;
    for ( int i = 0; i < hb_fw.EdgeObjectCount( ); i++ )
        x_fw.push_back( hb_fw.EdgeObject(i) );
    for ( int i = 0; i < hb_rc.EdgeObjectCount( ); i++ )
        x_rc.push_back( hb_rc.EdgeObject(i) );
    VecIntPairVec locs_fw, locs_rc;
    CreateGlocs( x_fw, L, locs_fw );
    CreateGlocs( x_rc, L, locs_rc );

    vec< vec<read_place> > PLACES2( bases.size( ) );
    const int min_qual = 1;
    const double prox = 50 * 1000;
    uint64_t nTerminated=0;
    if( bSafeFindPlaces ){
        std::cout << "being safe" << std::endl;
    }
    else{
        std::cout << "being unsafe" << std::endl;
    }

    #pragma omp parallel for schedule (dynamic, 1) reduction(+:nTerminated)
    for (size_t id = 0; id < bases.size(); id++) {
        int n = KmerId( bases[id], L, 0 );
        int qual_sum = infinity;
        if( bSafeFindPlaces ){
            const uint64_t maxNSteps=5000000;
            auto tmp = SafeFindPlaces( bases[id], quals[id], n, hb_fw, hb_rc, to_right_fw,
                                     to_right_rc, locs_fw, locs_rc, PLACES2[id], qual_sum,
                                     maxNSteps,
                                     min_qual, prox );
            if(tmp>0){
                ++nTerminated;
            }
        }
        else{
            FindPlaces( bases[id], quals[id], n, hb_fw, hb_rc, to_right_fw,
               to_right_rc, locs_fw, locs_rc, PLACES2[id], qual_sum, min_qual, prox );
        }
    }

    if( nTerminated > 0){
        std::cout << "\nWARNING: " + ToString(nTerminated) + "/" + ToString(bases.size())
                  + " reads have not been aligned to the K=1 bubble graph to avoid exponentially long run time.\n" << std::endl;
    }

    // Find out the best placement for each edge
    if (verbosity >= 1)
        cout << Date() << ": Finding the best placement for each read " << endl;

    vec<int> best_qsums(bubble_graph.EdgeObjectCount(), infinity);
    #pragma omp parallel
    {
        vec<int> best_qsums_loc(best_qsums);
        #pragma omp barrier
        #pragma omp for schedule (dynamic, 1)
        for (size_t id = 0; id < PLACES2.size(); id++) 
            for (int j = 0; j < PLACES2[id].isize(); j++)
                for (auto e: PLACES2[id][j].E()) 
                    best_qsums_loc[e] = min(best_qsums_loc[e], PLACES2[id][j].Qsum());
        #pragma omp critical
        {
            for( size_t edge=0;edge<best_qsums.size();++edge){
                best_qsums[edge] = min(best_qsums_loc[edge], best_qsums[edge]);
            }
        }
    }

    if (verbosity>=2) {
        const int MaxPlacementToDisplay = 1000;
        for (size_t id = 0; id < bases.size(); id++ ) {   
            cout << "read_" << id << endl;
            for ( int j = 0; j < PLACES2[id].isize( ); j++ ) {
                if (j >= MaxPlacementToDisplay) {
                    cout << "... " << PLACES2[id].isize() - j 
                        << " more ignored ..." << endl;
                    break;
                }
                vec<int> v;
                for ( int l = 0; l < PLACES2[id][j].N( ); l++ )
                    v.push_back( PLACES2[id][j].E(l) );
                cout << "    p=" << PLACES2[id][j].P() 
                    << " q= " <<  PLACES2[id][j].Qsum() << " ";
                v.Println(cout);
            }
        }
    }

    // Find the home of the reads, and quality score difference between the
    // best placement on this edge or in other edge
    // { { (edge11, diff11), (edge12, diff12), ...} // for read1
    //   { (edge21, diff21), (edge22, diff22), ...} // for read2
    //   ...}

    if (verbosity >= 1)
        cout << Date() << ": Finding the read support and qsum difference " << endl;

    vec< vec< pair<int,double>>> homes( bases.size( ) );
    vec<int> ids(bases.size(), vec<int>::IDENTITY);
    vec<int> nplaces(bases.size(), 0);
    for (size_t i = 0; i < PLACES2.size(); i++) 
        nplaces[i] = PLACES2[i].size();
    ReverseSortSync(nplaces, ids);
    
    #pragma omp parallel for schedule (dynamic, 1)
    for (size_t idx = 0; idx < ids.size(); idx++ ) {   
        int id = ids[idx];
        vec<int> v;
        for ( int j = 0; j < PLACES2[id].isize( ); j++ )
            for ( int l = 0; l < PLACES2[id][j].N( ); l++ )
                v.push_back( PLACES2[id][j].E(l) );
        UniqueSort(v);
        for ( int i = 0; i < v.isize( ); i++ ) {
            int e = v[i];
            bool to_show_read_support = Member(edges_to_show_supports, e);
            // what are the sister branches?
            vec<int> e2s;
            int n_left = to_left[e];
            for (size_t k = 0; k < bubble_graph.From(n_left).size(); k++) {
                int e2 = bubble_graph.EdgeObjectIndexByIndexFrom(n_left, k);
                e2s.push_back(e2);
            }
            int group_best_qsum = best_qsums[e];
            for (int e2: e2s) group_best_qsum = min(group_best_qsum, best_qsums[e2]);

            int q1 = infinity, q2 = infinity;
            int best_placement = -1;
            for ( int j = 0; j < PLACES2[id].isize( ); j++ ) {
                int qsum = PLACES2[id][j].Qsum();
                for (int e2: e2s) {
                    if (qsum > group_best_qsum + prox) continue;
                    if ( Member(PLACES2[id][j].E( ), e2) ) {
                        if (e2 == e) {
                            if (qsum < q1) {
                                q1 = qsum;
                                if (to_show_read_support) 
                                    best_placement = j;
                            }
                        }
                        else
                            q2 = Min( q2, qsum );
                    }
                }
            }
            if (q1 == infinity) continue;
            if (q2 == infinity) q2 = q1 + 30000; // +30 for single placement
            if (q1 < q2) {
                homes[id].push( e, q2 - q1 );    
                if (best_placement != -1) {
                    #pragma omp critical
                    edge_rid_place->push(e, id, PLACES2[id][best_placement]);
                }
            }
        }    
    }

    homes_index->clear_and_resize(bubble_graph.EdgeObjectCount());
    for (size_t id = 0; id < bases.size(); id++)
        for (int j = 0; j < homes[id].isize( ); j++)
            (*homes_index)[ homes[id][j].first ].push(id, homes[id][j].second);

    if (verbosity>=2) {
        vec<int> nreads( bubble_graph.EdgeObjectCount(), 0);
        for (size_t id = 0; id < bases.size(); id++ ) 
            for ( int j = 0; j < PLACES2[id].isize( ); j++ )
                for ( int l = 0; l < PLACES2[id][j].N( ); l++ )
                    nreads[ PLACES2[id][j].E(l) ]++;
        cout << "read support for each edges " << endl;
        for (size_t e = 0; e < nreads.size(); e++) {
            cout << "edge" << e << " nreads=" << nreads[e] << ": ";
            for (auto& x: (*homes_index)[e])
                cout << x.first << "(" << x.second << ")" << " ";
            cout << endl;
        }
    }
}

} // end namespace

struct SubEdgeLoc {
    int group_id;
    int branch_id;
    int shift;
    bool dir;
};

// Given the variants and callers, find the multiple placement of the caller
// edges and genome location (gid, pos) of the alternative placement.
void FindVariantFriends(const vec<VariantCallGroup>& vcall_groups, 
        const vec<vec<align>>& all_aligns, const HyperBasevector& hbp,
        const vec<pair<int,Bool>>& hbp_to_hb,
        map<Variant, vec<pair<int,int>>> *p_var_friending)
{
    map<int, vec<SubEdgeLoc>> edge_homes;
    for (size_t grpid = 0; grpid < vcall_groups.size(); grpid++) {
        const VariantCallGroup group = vcall_groups[grpid];
        for (int branch_id = 0; branch_id < group.GetNBranch(); branch_id++ ) {
            const vec<int>& path = group.GetBranchPath(branch_id);
            int shift = -group.GetTrim(branch_id);
            for (size_t sub_id = 0; sub_id < path.size(); sub_id++) {
                int edge0 = hbp_to_hb[path[sub_id]].first;
                bool dir  = hbp_to_hb[path[sub_id]].second;
                edge_homes[edge0].push_back(SubEdgeLoc{(int)grpid, branch_id, shift, dir});
                shift += hbp.EdgeLengthKmers(path[sub_id]);
            }
        }
    }

    // (grpid, caller) pair for each variant and (grpid, vcall) pair for each
    map<Variant, vec<pair<int,Caller>>> vars; 
    for (size_t grpid = 0; grpid < vcall_groups.size(); grpid++) {
        const VariantCallGroup group = vcall_groups[grpid];
        for (int i = 0; i < group.GetNBranch(); i++) {
            const vec<VariantCall>& vcall_vec = group.GetVariantCalls(i);
            for (const VariantCall& x: vcall_vec) {
                int edge0 = hbp_to_hb[x.caller.sub_edge_id].first;
                vars[x.variant].push(grpid, x.caller);
            }
        }
    }
    for (auto it = vars.begin(); it != vars.end(); ++it) {
        const Variant& var = it->first;
        // now assign friend locations
        for (const pair<int,Caller>& x: it->second) {
            int sub_edge = x.second.sub_edge_id;
            int sub_pos  = x.second.sub_edge_pos;
            int edge0    = hbp_to_hb[sub_edge].first;
            bool dir     = hbp_to_hb[sub_edge].second;
            int edge_len = hbp.EdgeLengthBases(sub_edge);

            set<int> grpids; // other groups this sub-edge belongs to
            for (SubEdgeLoc& loc: edge_homes[edge0]) { grpids.insert(loc.group_id); }
            if (grpids.size() <= 1) continue;

            int grpid = x.first;
            for (const SubEdgeLoc& loc: edge_homes[edge0]) { 
                if (loc.group_id == grpid) continue;
                int pos_friend_edge = (loc.dir == dir ? sub_pos : edge_len - 1 - sub_pos);
                int gpos = CorrelatePositionsRangeChecked(
                        vcall_groups[loc.group_id].GetAlign(loc.branch_id),
                        loc.shift + pos_friend_edge );
                if (gpos >= 0)
                    (*p_var_friending)[var].push(vcall_groups[loc.group_id].GetGID(), gpos);
            }
        }
    }
    for (auto it = (*p_var_friending).begin(); it != (*p_var_friending).end(); ++it) {
        UniqueSort(it->second);
    }
}


namespace {
    void PrintReadSupports(const HyperBasevector& bubble_graph,
            const vecbasevector& bases,
            const vec<int> sample_ids,
            vec<tuple<int,int,read_place>> edge_rid_place) 
    {
        if (edge_rid_place.empty()) return;
        vec<int> to_left; bubble_graph.ToLeft(to_left);

        auto CompareByEdgeSampleRead = 
            [&sample_ids](const tuple<int,int,read_place>& a, const tuple<int,int,read_place>&b)
            {   
                int e, rid, e2, rid2;
                tie(e, rid, ignore) = a;
                tie(e2, rid2, ignore) = b;
                int sample = sample_ids[rid];
                int sample2 = sample_ids[rid2];
                return tie(e, sample, rid) < tie(e2, sample2, rid2);
            };
        sort(edge_rid_place.begin(), edge_rid_place.end(),
                CompareByEdgeSampleRead);

        vec<int> shifts(edge_rid_place.size(), 0);
        for (size_t i = 0; i < edge_rid_place.size(); i++) {
            int e = -1, rid = -1;
            read_place rp;
            tie(e, rid, rp) = edge_rid_place[i];
            vec<int> rshifts;
            if (!rp.Fw()) rp.Reverse(bases[rid], bubble_graph);
            rp.FindReadShift(e, bubble_graph, rshifts);
            shifts[i] = rshifts[0];
        }

        int min_shift = *min_element(shifts.begin(), shifts.end());

        TextTable tb;
        for (size_t i = 0; i < edge_rid_place.size(); i++) {
            int e = get<0>(edge_rid_place[i]);
            size_t j = i+1;
            while (j < edge_rid_place.size() && 
                    get<0>(edge_rid_place[j]) == e ) j++;

            tb.SetRawLine();
            tb << " ======================================== Edge " << e 
                << " ======================================= " << EndRow;

            // the reference segments +/- 100
            const int VarRefExtMax = 100;
            int node_left = to_left[e];
            int e_ref = bubble_graph.EdgeObjectIndexByIndexFrom(node_left,
                    bubble_graph.FromSize(node_left)-1);
            basevector ref_seg;
            int ext_len = 0;
            ExtendedBasevector(bubble_graph, e_ref, min(VarRefExtMax, -min_shift), 
                    ref_seg , ext_len);
            ForceAssertLe(ext_len, -min_shift);

            tb << "REF extended" << Tab << String(-ext_len-min_shift, ' ') << ref_seg << EndRow;
            // the variant 
            tb << "ALT" << Tab << String(-min_shift, ' ') << bubble_graph.EdgeObject(e) << EndRow;
            tb << "REF" << Tab << String(-min_shift, ' ') << bubble_graph.EdgeObject(e_ref) << EndRow;
            // the reads
            for (size_t k = i; k < j; k++) {
                int rid = get<1>(edge_rid_place[k]);
                const read_place& rp = get<2>(edge_rid_place[k]);
                basevector read = bases[rid];
                int sample = sample_ids[rid];
                if (!rp.Fw()) 
                    read.ReverseComplement();
                tb << sample << ":" << rid << (rp.Fw() ? "+":"-" ) 
                    << " q " << rp.Qsum()/1000 << Tab;
                tb << String(shifts[k] - min_shift, ' ') << read << EndRow;
            }
            i = j - 1;
        }
        tb.Print(cout, 1, "ll");
    }
}

// Build a K=1 hyperbasevector of the reference with bubbles for each variant,
// or several variants if they are in the same ref position. Then calculate the
// probability of the variants by aligning reads back to it.
// The output is a map of each variant to a vector of pairs:
// {   {prob_ref, prob_var} .. from reads in sample0
//     {prob_ref, prob_var} .. from reads in sample1
//     .... }
void FindVariantProb(const ReadOriginTracker* p_read_tracker,
        const vec<VariantCallGroup>& vcall_groups,
        const vecbasevector& Gplus, 
        const vec<int>& Gplus_ext, 
        map<Variant, vec<pair<double,double>>>& probs,
        vec<int> vids_to_show_supports,
        int verbosity,const bool bSafeFindPlaces) 
{
    if (verbosity >= 1) 
        cout << Date() << ": Finding probabilities for each variant" << endl;
    
    const vecbasevector& bases = p_read_tracker->Reads();
    const vecqualvector& quals = p_read_tracker->Quals();
    int nsamples = p_read_tracker->getSampleList().size();

    set<Variant> varset;
    for (const auto& group: vcall_groups)
        for( const auto& vcall_vec: group.GetVariantCalls())
            for (const VariantCall& x: vcall_vec) 
                varset.insert(x.variant); 
    if (varset.empty()) return;
    vec<Variant> vars(varset.begin(), varset.end());

    vec<vec<vec<basevector>>> vars_seq(Gplus.size());
    vec<vec<vec<int>>> vars_index(Gplus.size());
    vec<vec<pair<int,int>>> bound(Gplus.size());
    for (size_t i = 0; i < vars.size(); i++) {
        int g = vars[i].gid;
        int start = vars[i].pos;
        int end = start + vars[i].ref.size();
        size_t i2 = i+1;
        while (i2 < vars.size() && vars[i2].gid == g && vars[i2].pos < end) {
            end = max(end, vars[i2].pos + (int)vars[i2].ref.size());
            i2++;
        }
        start += Gplus_ext[g];
        end += Gplus_ext[g];
        bound[g].push(start, end);
        vars_seq[g].push_back(vec<basevector>());
        vars_index[g].push_back(vec<int>());
        for(size_t k = i; k < i2; k++) {
            int start_v= vars[k].pos + Gplus_ext[g];
            int len = vars[k].ref.size();
            basevector b1(Gplus[g], start, start_v - start);
            //basevector b2(Gplus[g], start_v, len); //ref
            basevector b2(vars[k].alt);
            basevector b3(Gplus[g], start_v + len, end - start_v - len);
            vars_seq[g].back().push_back(Cat(b1,b2,b3));
            vars_index[g].back().push_back(k);
        }
        i = i2 - 1;
    }
    HyperBasevector variant_graph(1);
    map<int,int>  edge_to_varid;
    for (size_t g = 0; g < vars_seq.size(); g++) {
        if (vars_seq[g].empty()) continue;
        variant_graph.AddVertices(1);
        // add precedding edge if necessary
        if (bound[g].front().first > 0) { 
            basevector edge(Gplus[g], 0, bound[g].front().first);
            variant_graph.AddVertices(1);
            variant_graph.AddEdge(variant_graph.N()-2, variant_graph.N()-1, edge);
            edge_to_varid[variant_graph.EdgeObjectCount()-1] = -1;
        }
        for (size_t i = 0; i < vars_seq[g].size(); i++) {
            variant_graph.AddVertices(1);
            for (size_t j = 0; j < vars_seq[g][i].size(); j++) {
                variant_graph.AddEdge(variant_graph.N()-2, variant_graph.N()-1, 
                        vars_seq[g][i][j]);
                edge_to_varid[variant_graph.EdgeObjectCount()-1] = vars_index[g][i][j];
            }
            // add the ref edge for the variants
            basevector ref_edge(Gplus[g], bound[g][i].first, bound[g][i].second - bound[g][i].first);
            variant_graph.AddEdge(variant_graph.N()-2, variant_graph.N()-1, ref_edge);
            edge_to_varid[variant_graph.EdgeObjectCount()-1] = -1;
            // add the ref edge between variants
            if (i < vars_seq[g].size()-1 && bound[g][i+1].first > bound[g][i].second ){
                //cout << "add edge from " << bound[g][i].second 
                //    << " to " << bound[g][i+1].first << " len = " 
                //    << bound[g][i+1].first - bound[g][i].second <<  endl;
                basevector ref_edge(Gplus[g],  bound[g][i].second, 
                        bound[g][i+1].first - bound[g][i].second);
                variant_graph.AddVertices(1);
                variant_graph.AddEdge(variant_graph.N()-2, variant_graph.N()-1, ref_edge);
                edge_to_varid[variant_graph.EdgeObjectCount()-1] = -1;
            }
        }
        // add trailing edge if necessary
        if (bound[g].back().second < (int)Gplus[g].size()) {
            basevector edge(Gplus[g].begin()+bound[g].back().second, Gplus[g].end());
            variant_graph.AddVertices(1);
            variant_graph.AddEdge(variant_graph.N()-2, variant_graph.N()-1, edge);
            edge_to_varid[variant_graph.EdgeObjectCount()-1] = -1;
        }
    }
    if (verbosity >= 1) 
        DumpGraph("vgroup.dot", variant_graph);

    if (verbosity >= 3)
        for (int e = 0; e < variant_graph.EdgeObjectCount(); e++) {
            cout << "e= " << e << " vid= " << edge_to_varid[e]
                << " : " << variant_graph.EdgeObject(e) << endl;
        }

    vec<int> edges_to_show_supports(vids_to_show_supports.size());
    if (!vids_to_show_supports.empty()) {
        for (size_t i = 0; i < vids_to_show_supports.size(); i++) 
            for (int e = 0; e < variant_graph.EdgeObjectCount(); e++)
                if (vids_to_show_supports[i] == edge_to_varid[e])
                    edges_to_show_supports[i] = e;
        cout << "Looking for support for the following variants" << endl;
        for (size_t i = 0; i < edges_to_show_supports.size(); i++) {
            cout << "vid= " << vids_to_show_supports[i] 
                << " edge= " << edges_to_show_supports[i] << endl;
        }
        // Always add the ref edge for each variant
        vec<int> edges_to_show_supports2;
        vec<int> to_left; variant_graph.ToLeft(to_left);
        for (int e: edges_to_show_supports) {
            int node_left = to_left[e];
            edges_to_show_supports2.push_back(e);
            for (int i = 0; i < variant_graph.FromSize(node_left); i++)
                edges_to_show_supports2.push_back(variant_graph.
                        EdgeObjectIndexByIndexFrom(node_left,i));
        }
        swap(edges_to_show_supports2, edges_to_show_supports);
    }

    vec< vec< pair<int,int> > > home_index; // pair of (rid, qsum)
    vec<tuple<int,int,read_place>> edge_rid_place;
    FindReadHomesBest(bases, quals, variant_graph, &home_index, 
            edges_to_show_supports, &edge_rid_place,
            verbosity,bSafeFindPlaces);

    if (!edge_rid_place.empty()) {
        vec<int> sample_ids(bases.size(), -1);
        for (size_t i = 0; i < sample_ids.size(); i++) 
            sample_ids[i] = p_read_tracker->getSampleID(i);
        PrintReadSupports(variant_graph, bases, sample_ids, edge_rid_place);
    }

    for (int n1 = 0; n1 < variant_graph.N(); n1++) {
        if (variant_graph.From(n1).size() <= 1) continue;
        vec<int> branches(variant_graph.From(n1).size());
        for (size_t j = 0; j < variant_graph.From(n1).size(); j++) 
            branches[j] = variant_graph.EdgeObjectIndexByIndexFrom(n1,j);
        size_t nbranches = branches.size();
        const int qgood = 4;
        // the reference branch is the last one
        vec<vec<double>> qsum(nbranches, vec<double>(nsamples, 0.0));
        vec<vec<int>> nreads(nbranches, vec<int>(nsamples, 0));
        for (size_t j = 0; j < branches.size(); j++) {
            int e = branches[j];
            for (auto& x: home_index[e]) {
                int rid = x.first;
                int sample_id = p_read_tracker->getSampleID(rid);
                if (x.second/1000 >= qgood && sample_id >= 0) {
                    qsum[j][sample_id] += x.second/1000;
                    nreads[j][sample_id] ++;
                    if (verbosity >= 2) {
                        cout << "add read_" << rid << " (" 
                            << sample_id << ") to edge " << e 
                            << " qsum= " << x.second/1000 << endl;
                    }
                }
            }
        }
        if (verbosity >= 2) {
            cout << "branch: " << endl;
            for (size_t j = 0; j < branches.size(); j++) {
                int e = branches[j];
                if (j == branches.size() -1) 
                    cout << "reference ";
                else
                    cout << "vid = " << edge_to_varid[e] + 1;
                cout << " e= " << e;
                for (int sample = 0; sample < nsamples; sample++)
                    cout << " nreads= " << nreads[j][sample] << " q= " << qsum[j][sample];
                cout  << endl;
            }
        }
        // Doe not call if there are nread <=2 && q <= 40 while all other
        // branches have 5 times ore more qsum.
        const int WeakSupport = 2, WeakQSum = 40;
        const int StrongQsumFactor = 5;
        for (int sample = 0; sample < nsamples; sample++) {
            vec<int> weak;
            double max_weak_support = 0;
            for (size_t j = 0; j < branches.size(); j++) {
                if (nreads[j][sample] > 0 && nreads[j][sample] <= WeakSupport &&
                        qsum[j][sample] <= WeakQSum) {
                    weak.push_back(j);
                    max_weak_support = max(max_weak_support, qsum[j][sample]);
                }
            }
            if (weak.empty()) continue;
            size_t ngood = 0, nzero = 0;
            for (size_t j = 0; j < branches.size(); j++) 
                if (qsum[j][sample] >= StrongQsumFactor * max_weak_support)
                    ngood++;
                else if (nreads[j][sample] == 0) 
                    nzero++;
            if (weak.size() + ngood + nzero == branches.size() && ngood > 0) {
                for (int j: weak) {
                    qsum[j][sample] = 0.;
                    if (verbosity >= 2 && nreads[j][sample] != 0)
                        cout << "Disable branch " << j << " for sample " << sample
                            << " vid = " << edge_to_varid[branches[j]] + 1 
                            << " nread= " << nreads[j][sample]
                            << " qsum= " << qsum[j][sample]
                            << endl;
                }
            }
        }
        for (size_t j = 0; j < branches.size()-1; j++) {
            int vid = edge_to_varid[branches[j]];
            ForceAssertNe(vid, -1);
            for (int sample = 0; sample < nsamples; sample++) 
                probs[vars[vid]].push(qsum[nbranches-1][sample], qsum[j][sample]);
        }
    }
}

// ======================= EdgesOnRef method implementation ====================================

void EdgesOnRef::InitFromBestPath(const vec<int>& path, 
        const vec<pair<int,int>>& ref_pos) 
{
    ForceAssert(dg_.N() == 0);
    vec<int> to_right; hb_.ToRight(to_right);
    vec<int> to_left; hb_.ToLeft(to_left);
    dg_.AddVertices(1); // the first vertex
    for (size_t i = 0; i < path.size(); i++) {
        // additional vertex if two edges are not connected in assembly graph
        if (i > 0 && to_right[path[i-1]] != to_left[path[i]])
            dg_.AddVertices(1);
        dg_.AddVertices(1);
        dg_.AddEdge(dg_.N()-2, dg_.N()-1, EdgeLoc(path[i], 
                    ref_pos[i].first + gplus_ext_));
    }
}

// Unroll the graph, find alignments of the unrolled edges, find variants,
// and return the path of the edges.
void EdgesOnRef::UnrollAll(const int verbosity, const int iLoBound, const int iHiBound, const bool bCorrection)
{
    std::unordered_set<int> prefered_edges;
    for(int ii=0;ii<dg_.EdgeObjectCount();++ii){
        prefered_edges.insert(ii);
    }
    if (verbosity >= 1) {
        cout << Date() << ": Start unrolling using best paths" << endl;
        cout << "    best path= ";
        for(int i = 0; i < dg_.EdgeObjectCount(); i++)
            cout << GetOriginalId(i) << " ";
        cout << endl;
        std::cout <<"    unroll limit: " << iLoBound << " " <<iHiBound<<std::endl;
        DumpGraph("hbp.dot", hb_);
    }
    ForceAssertGt(dg_.EdgeObjectCount(), 0);
    const int LenMax = 2000;   // for graph traversal

    vec<int> to_right; hb_.ToRight(to_right);
    vec<int> to_left; hb_.ToLeft(to_left);
    vec<int> backbone(dg_.EdgeObjectCount(), vec<int>::IDENTITY);
    const int MinOverlap = 80;

    if (verbosity >= 1) cout << Date() << ": Extending the path tail" << endl;
    int tail_edge = dg_.EdgeObjectCount() - 1;
    while (true) {
        int tail_edge0 = GetOriginalId(tail_edge);
        int tail_edge_end = hb_.EdgeLengthKmers(tail_edge0) + GetStart(tail_edge);
        if (verbosity >= 2) 
            cout << "Trailing edge " << tail_edge << "->" << tail_edge0 << endl;
        if (hb_.From(to_right[tail_edge0]).empty()) break;

        int new_edge0 = hb_.EdgeObjectIndexByIndexFrom(to_right[tail_edge0], 0);
        bool is_loop = false;
        for (auto& x: backbone) {
            if (GetOriginalId(x) == new_edge0) {
                is_loop = true;
                break;
            }
        }
        if (is_loop) {
            if (verbosity >= 2) 
                cout << "loop encountered" << endl;
            break;
        }
        align a;
        int nerror;
        int Bandwidth = max(1, (int)hb_.EdgeObject(new_edge0).size()/2);
        SmithWatAffineBanded(hb_.EdgeObject(new_edge0), gplus_, -tail_edge_end, Bandwidth, a, nerror);
        if (verbosity >= 2) {
            cout << "align at " << a.pos1() << "," << a.Pos1() << " to " << a.pos2() << "," << a.Pos2() << endl;
            PrintVisualAlignment(True, cout, hb_.EdgeObject(new_edge0), gplus_, a);
        }
        int new_edge_start = a.pos2() - a.pos1();
        if (  new_edge_start + MinOverlap > (int)gplus_.size()
            || a.pos2() > iHiBound || a.Pos2() > iHiBound
           ) break;

        if (verbosity >= 2) 
            cout << "Adding tail " << new_edge0 << " start " << new_edge_start << endl;
        vec<pair<int,int>> tail_extension;
        tail_extension.push(new_edge0, new_edge_start );
        AddConnections(tail_edge, -1, tail_extension);
        tail_edge = dg_.EdgeObjectCount() - 1;
        backbone.push_back(tail_edge);
    }

    if (verbosity >= 1) cout << Date() << ": Extending the path head" << endl;
    int head_edge = 0;
    while (true) {
        int head_edge0 = GetOriginalId(head_edge);
        int start0 = GetStart(head_edge);
        if (verbosity >= 2) 
            cout << "head edge " << head_edge0 << " start at " << start0 << endl;
        if(hb_.To(to_left[head_edge0]).empty()) break;

        int new_edge0 = hb_.EdgeObjectIndexByIndexTo(to_left[head_edge0], 0);
        bool is_loop = false;
        for (auto& x: backbone) {
            if (GetOriginalId(x) == new_edge0) {
                is_loop = true;
                break;
            }
        }
        if (is_loop) {
            if (verbosity >= 2) 
                cout << "loop encountered" << endl;
            break;
        }
        int start_prev = start0 - hb_.EdgeLengthKmers(new_edge0);
        align a;
        int nerror;
        int Bandwidth = max(1, (int)hb_.EdgeObject(new_edge0).size()/2);
        SmithWatAffineBanded(hb_.EdgeObject(new_edge0), gplus_, -start_prev, Bandwidth, a, nerror);
        if (verbosity >= 2) {
            cout << "align at " << a.pos1() << "," << a.Pos1() << " to " << a.pos2() << "," << a.Pos2() << endl;
            PrintVisualAlignment(True, cout, hb_.EdgeObject(new_edge0), gplus_, a);
        }
        if (a.Pos2() < MinOverlap
            || a.pos2() < iLoBound || a.Pos2() < iLoBound
           ) break;

        int new_edge_start = a.pos2() - a.pos1();
        vec<pair<int,int>> head_extension;
        head_extension.push(new_edge0, new_edge_start);
        if (verbosity >= 2) 
            cout << "Adding head " << new_edge0 << " starting at " << new_edge_start << endl;
        AddConnections(-1, head_edge, head_extension);
        head_edge = dg_.EdgeObjectCount() - 1;
        backbone.push_front(head_edge);
    }
    vec<int> backbone_dup(backbone.size(), 0);
    if (verbosity >= 2) {
        cout << "The backbone edges are";
        backbone.Println(cout);
    }
    // each edge belongs to a path between two edges [e1,.., e2], where e1
    // and e2 are all in the backbone
    map<int, pair<int,int>> bb_anchors; 
    for (auto& x: backbone) 
        bb_anchors[x] = make_pair(x,x);



    if (verbosity >= 1) 
        cout << Date() << ": Looking for alternative paths between backbone edges" << endl;
    const int MaxLocDevAllowed = 150;
    while (true) {
        bool no_change = true; // keep searching if alternative path is found
        vec<vec<pair<int,int>>> known_starts(hb_.EdgeObjectCount());
        for (int e2 = 0; e2 < dg_.EdgeObjectCount(); e2++) 
            known_starts[GetOriginalId(e2)].push(GetStart(e2), e2);

        for (int node2 = 0; node2 < dg_.N() && no_change; node2++) {
            for (size_t j = 0; j < dg_.From(node2).size(); j++) {
                // For each edge
                int edge2 = dg_.EdgeObjectIndexByIndexFrom(node2, j);
                int edge1 = GetOriginalId(edge2);
                int node2_next = dg_.From(node2)[j];
                int node1_next = to_right[edge1];
                // start from node1_next in graph1 and find a new path from
                // known paths starting from node_next in graph2 
                set<int> known_edges1;
                for (size_t k = 0; k < dg_.From(node2_next).size(); k++) {
                    int edge2_next = dg_.EdgeObjectIndexByIndexFrom(node2_next, k);
                    known_edges1.insert(GetOriginalId(edge2_next));
                }
                for (size_t k = 0; k < hb_.From(node1_next).size() && no_change; k++) {
                    int edge1_next = hb_.EdgeObjectIndexByIndexFrom(node1_next, k);
                    int node1_nn = hb_.From(node1_next)[k];
                    if (find(known_edges1.begin(), known_edges1.end(), edge1_next)
                            != known_edges1.end()) continue;
                    if (verbosity >= 2)
                        cout << "try path "<< edge2 << "(" << edge1 << ") -> (" 
                            << edge1_next << ") " << endl;

                    // Recursively follow the edge until reaching in backbone again
                    // search for possible alternative paths from edge1 to back_edge1:
                    //    (xx) --edge1--> (node1_next) --edge1_next--> (node1_nn) --back_edge1--
                    // (node2) --edge2--> (node2_next) --edge2_next--> (xx)       --back_edge2--
                    int back_edge = -1;
                    vec<pair<int,int>> alternative_path_and_start; 
                    int start_next = GetStart(edge2) + hb_.EdgeLengthKmers(edge1);
                    alternative_path_and_start.push(edge1_next, start_next);
                    if (hb_.To(node1_next).solo() && hb_.From(node1_nn).solo() &&
                            known_edges1.size() == 1 && 
                            to_right[*known_edges1.begin()] == node1_nn) { // alway add simple bubble
                        int edge1_nn = hb_.EdgeObjectIndexByIndexFrom(node1_nn, 0);
                        if (known_starts[edge1_nn].solo()) {
                            back_edge = known_starts[edge1_nn][0].second;
                        }
                    }
                    if (back_edge == -1) {
                        FindAlternativePath(node1_nn, &alternative_path_and_start, &back_edge, 
                                known_starts, LenMax, MaxLocDevAllowed, verbosity);
                    }
                    if (back_edge != -1) {
                        if (verbosity >= 2)
                        {
                            cout << "found alternative path between " << edge2 << "(" << edge1 << ")" << " and " 
                                << back_edge << "(" << GetOriginalId(back_edge) << ")" << endl;
                            for (auto&x: alternative_path_and_start) 
                                cout << x.first << "@" << x.second << " ";
                            cout << endl;
                        }
                        int new_edge_start = dg_.EdgeObjectCount();
                        AddConnections(edge2, back_edge, alternative_path_and_start);
                        int new_edge_end = dg_.EdgeObjectCount();

                        int anchor_start = bb_anchors[edge2].first;
                        int anchor_end = bb_anchors[back_edge].second;
                        for (int x = new_edge_start; x <= new_edge_end; x++) 
                            bb_anchors[x] = make_pair(anchor_start, anchor_end);
                        int pos1 = find(backbone.begin(), backbone.end(), anchor_start) - backbone.begin();
                        int pos2 = find(backbone.begin(), backbone.end(), anchor_end) - backbone.begin();
                        ForceAssertNe(pos1, backbone.isize());
                        ForceAssertNe(pos2, backbone.isize());
                        if (pos2 > pos1) {
                            for (int x = pos1+1; x < pos2; x++)
                                backbone_dup[x]++;
                        } else { // loop back
                            for (int x = pos2; x <= pos1; x++)
                                backbone_dup[x]++;
                        }
                        no_change = false;
                    }
                }
            }
        }
        if (no_change) break;
    }


    if(verbosity >= 2){
        std::cout<< "prefered edges ";
        for( const auto& entry: prefered_edges){
            std::cout << entry << " " ;
        }
        std::cout<< std::endl;
    }



for( bool bLoop=bCorrection;bLoop;)
{
    bLoop=false;
    auto dups = count_dg_dups();
    if( dups.size() > 0){
        vec<int> dg_to_right; dg_.ToRight(dg_to_right);
        vec<int> dg_to_left; dg_.ToLeft(dg_to_left);
        for( auto& entry: dups){
            std::cout << std::get<0>(entry) << " "
                      << std::get<1>(entry) << " "
                      << std::get<2>(entry) << " "
                      << std::get<3>(entry) << " "
                      << std::endl;
        }
        auto connection_counts = count_dg_prefered_edge_connection(prefered_edges,dg_to_left,dg_to_right);
//        std::cout <<"connection count: "<< std::endl;
//        for(size_t ii=0;ii<connection_counts.size();++ii){
//            std::cout << ii << " " << connection_counts[ii].first << " " << connection_counts[ii].second << std::endl;
//        }
        vec<std::tuple<size_t,int,size_t,bool,int,int,int,int>> dups_counted(dups.size());
        for(size_t ii=0;ii<dups.size();++ii){
            const auto& entry= dups[ii];
            auto vertex = std::get<0>(entry) ;
            auto vi = std::get<1>(entry) ;
            auto vj = std::get<2>(entry) ;
            auto dir = std::get<3>(entry) ;

            if( dir==0){ // to
                auto ei = dg_.EdgeObjectIndexByIndexTo(vertex,vi);
                auto ej = dg_.EdgeObjectIndexByIndexTo(vertex,vj);
                ForceAssert( dg_.EdgeObject(ei).original_edge == dg_.EdgeObject(ej).original_edge);

                size_t ci = connection_counts[ei].first;
                size_t cj = connection_counts[ej].first;
                bool bji = cj>ci;
                int loser = bji?ei:ej;
                dups_counted[ii]=std::make_tuple(max(ci,cj),loser,min(ci,cj),bji,vertex,vi,vj,dir);
            }
            else{ // from
                auto ei = dg_.EdgeObjectIndexByIndexFrom(vertex,vi);
                auto ej = dg_.EdgeObjectIndexByIndexFrom(vertex,vj);
                ForceAssert( dg_.EdgeObject(ei).original_edge == dg_.EdgeObject(ej).original_edge);

                size_t ci = connection_counts[ei].second;
                size_t cj = connection_counts[ej].second;
                bool bji = cj>ci;
                int loser = bji?ei:ej;
                dups_counted[ii]=std::make_tuple(max(ci,cj),loser,min(ci,cj),bji,vertex,vi,vj,dir);
            }
        }
        Sort(dups_counted);
        for( auto& entry: dups_counted){
            auto c1 = std::get<0>(entry) ;
            auto loser = std::get<1>(entry);
            auto c2 = std::get<2>(entry) ;
            auto bji = std::get<3>(entry) ;
            auto vertex = std::get<4>(entry) ;
            auto vi = std::get<5>(entry) ;
            auto vj = std::get<6>(entry) ;
            auto dir = std::get<7>(entry) ;

            auto ci = (bji)?c2:c1;
            auto cj = (bji)?c1:c2;

            int ei,ej;
            if( dir==0){ // to
                ei = dg_.EdgeObjectIndexByIndexTo(vertex,vi);
                ej = dg_.EdgeObjectIndexByIndexTo(vertex,vj);
            }
            else{
                ei = dg_.EdgeObjectIndexByIndexFrom(vertex,vi);
                ej = dg_.EdgeObjectIndexByIndexFrom(vertex,vj);
            }
            ForceAssert( dg_.EdgeObject(ei).original_edge == dg_.EdgeObject(ej).original_edge);
            std::cout << dg_.EdgeObject(ei).original_edge << " "
                      << dir << "|"
                      << loser << " "
                      << ei << " "
                      << ej << " "
                      << ci << " "
                      << cj << " "<< std::endl;
        }
        vec<int> to_delete;
        clear_dg_branch(to_delete, std::get<1>(dups_counted.back()),std::get<7>(dups_counted.back()),dg_to_left,dg_to_right  );
        Sort(to_delete);
        for(auto& entry: backbone){
            int count = 0;
            for(count=0;count<entry;++count){ }
            ForceAssert(count<=entry);
            entry-=count;
        }
        dg_.DeleteEdges(to_delete);
        bLoop=true;
    }
}

    if (verbosity >= 2)
    {
        cout << "Seeding edges are: " ;
        for (size_t i = 0; i < backbone.size(); i++) {
            if (backbone_dup[i] == 0) cout << backbone[i] << " ";
        }
        cout << endl;
    }
    backbone_ = backbone;
    backbone_dup_ = backbone_dup;
};


vec<std::tuple<int,int,int,int>> EdgesOnRef::count_dg_dups()const{
    vec<std::tuple<int,int,int,int>> out;
    for (int vertex = 0; vertex < dg_.N()  ; vertex++) {
        for (size_t ii = 0; ii < dg_.To(vertex).size(); ii++) {
            auto ei_i = dg_.EdgeObjectIndexByIndexTo(vertex,ii);
            for (size_t jj = ii+1; jj < dg_.To(vertex).size(); jj++) {
                auto ei_j = dg_.EdgeObjectIndexByIndexTo(vertex,jj);
                if( dg_.EdgeObject(ei_i).original_edge == dg_.EdgeObject(ei_j).original_edge){
                    std::cout <<  dg_.EdgeObject(ei_i).original_edge << std::endl;
                    out.push_back( std::make_tuple(vertex,ii,jj,0) );
                }
            }
        }
        for (size_t ii = 0; ii < dg_.From(vertex).size(); ii++) {
            auto ei_i = dg_.EdgeObjectIndexByIndexFrom(vertex,ii);
            for (size_t jj = ii+1; jj < dg_.From(vertex).size(); jj++) {
                auto ei_j = dg_.EdgeObjectIndexByIndexFrom(vertex,jj);
                if( dg_.EdgeObject(ei_i).original_edge == dg_.EdgeObject(ei_j).original_edge){
                    std::cout <<  dg_.EdgeObject(ei_i).original_edge << std::endl;
                    out.push_back( std::make_tuple(vertex,ii,jj,1) );
                }
            }
        }
    }
    return out;
}
vec<std::pair<size_t,size_t>> EdgesOnRef::count_dg_prefered_edge_connection(const std::unordered_set<int>& prefered_edges
                                                                           ,const vec<int>& to_left, const vec<int>& to_right)const{
    vec<std::pair<size_t,size_t>> out(dg_.EdgeObjectCount(),std::make_pair(0ul,0ul));
    for( auto pe: prefered_edges){
        //going right
        const size_t nBases = hb_.EdgeObject(GetOriginalId(pe)).size();
        {
            std::deque<int> to_do;
            std::set<int> discovered;
            to_do.push_back(pe);
            discovered.insert(pe);
            while(!to_do.empty()){
                auto edge=to_do.front();
                to_do.pop_front();

                out[edge].first+=nBases;
                auto w = to_right[edge];
                for( int ee=0 ; ee < dg_.FromSize(w);++ee){
                    auto ei = dg_.EdgeObjectIndexByIndexFrom(w,ee);
                    if( discovered.find(ei) == discovered.end()){
                        discovered.insert(ei);
                        to_do.push_back(ei);
                    }
                }
            }
        }
        //going left
        {
            std::deque<int> to_do;
            std::set<int> discovered;
            to_do.push_back(pe);
            discovered.insert(pe);
            while(!to_do.empty()){
                auto edge=to_do.front();
                to_do.pop_front();

                out[edge].second+=nBases;
                auto v = to_left[edge];
                for( int ee=0 ; ee < dg_.ToSize(v);++ee){
                    auto ei = dg_.EdgeObjectIndexByIndexTo(v,ee);
                    if( discovered.find(ei) == discovered.end()){
                        discovered.insert(ei);
                        to_do.push_back(ei);
                    }
                }
            }
        }
    }
    return out;
}

void EdgesOnRef::clear_dg_branch(vec<int>& to_delete, int root_edge,int dir,const vec<int>& to_left,const vec<int>&to_right){
    to_delete.clear();
    std::set<int> deleted;
    std::deque<int> to_do;
    to_do.push_back(root_edge);
    while(!to_do.empty()){
        int edge = to_do.front();
        to_do.pop_front();
        if( deleted.find(edge)== deleted.end()){
            deleted.insert(edge);
            to_delete.push_back(edge);


            size_t nParallelAway=0;
            if( dir == 0){
                int vertex = to_left[edge];
                for(int ee=0;ee<dg_.FromSize(vertex);++ee){
                    if( deleted.find(dg_.EdgeObjectIndexByIndexFrom(vertex,ee))==deleted.end()){
                        ++nParallelAway;
                    }
                }
            }
            else{
                int vertex = to_right[edge];
                for(int ee=0;ee<dg_.ToSize(vertex);++ee){
                    if( deleted.find(dg_.EdgeObjectIndexByIndexTo(vertex,ee))==deleted.end()){
                        ++nParallelAway;
                    }
                }
            }
            if(nParallelAway==0){
                if( dir == 0){
                    int vertex = to_left[edge];
                    for(int ee=0;ee<dg_.ToSize(vertex);++ee){
                        to_do.push_back( dg_.EdgeObjectIndexByIndexTo(vertex,ee));
                    }
                }
                else{
                    int vertex = to_right[edge];
                    for(int ee=0;ee<dg_.FromSize(vertex);++ee){
                        to_do.push_back( dg_.EdgeObjectIndexByIndexFrom(vertex,ee));
                    }
                }
            }
        }
    }
    Sort(to_delete,std::greater<int>());
    /*
    std::cout<<"deleting: ";
    for(auto entry:to_delete){
        std::cout << entry << " ";
    }
    std::cout<<std::endl;
    dg_.DeleteEdges(to_delete);
    */
}

// Starting from an edge, exploring all path until it lands to the unrolled graph and
// the inferred edge location is consistent with know value.
bool EdgesOnRef::FindAlternativePath(int node, vec<pair<int,int>> *p_edge_starts, int *p_back_edge, 
        const vec<vec<pair<int,int>>>& known_starts, 
        const int LenMax, const int MaxLocDevAllowed,
        int verbosity) 
{
    ForceAssertGt(p_edge_starts->size(), 0u);
    for (size_t i = 0; i < hb_.From(node).size(); i++) {
        int node_new = hb_.From(node)[i];
        int edge_new = hb_.EdgeObjectIndexByIndexFrom(node, i);
        int start_new = p_edge_starts->back().second + 
            hb_.EdgeLengthKmers(p_edge_starts->back().first) ;
        // do not allow loop
        bool is_loop = false;
        for (auto& x: *p_edge_starts) {
            if (x.second == edge_new) {
                is_loop = true;
                break;
            }
        }
        if (is_loop) continue;
        // check if it's a valid alternative path, either if the path end
        // shift agrees with known one, or the path belongs to a simple bubble
        const vec<pair<int,int>>& starts = known_starts[edge_new];
        *p_back_edge = -1; // find the tail edge?
        for (const auto& x : starts) {
            if (abs(x.first - start_new) < MaxLocDevAllowed) {
                *p_back_edge = x.second;
                break;
            }
        }
        if (*p_back_edge != -1) // path end agrees
            return true;
        // not found, track down
        if (start_new - p_edge_starts->front().second > LenMax
                || starts.size() > 0 || p_edge_starts->size() > 10) {
            if (starts.size() > 0 && verbosity >= 2) {
                cout << "skip mismatch edge  " << edge_new << " expect_start= " << start_new 
                    << " known_start= " ;
                for (const auto& x : starts) cout << x.first << " "; cout << endl;
            }
            continue;
        }
        p_edge_starts->push_back(make_pair(edge_new,start_new));
        //cout << "follow edge ->" << edge_new << endl;
        if (FindAlternativePath(node_new, p_edge_starts, p_back_edge, known_starts, 
                    LenMax, MaxLocDevAllowed, verbosity))
            return true;
    }
    // all branches failed, clear last node, track back.
    p_edge_starts->pop_back();
    return false;
}

void EdgesOnRef::AddConnections(int front_edge, int back_edge, const vec<pair<int,int>>& alternative_path_and_start)
{
    vec<int> to_right; dg_.ToRight(to_right);
    vec<int> to_left; dg_.ToLeft(to_left);
    // add the missing node if needed
    int last_node = (back_edge == -1 ? dg_.N() : to_left[back_edge]);
    if (back_edge == -1) dg_.AddVertices(1);
    int first_node = (front_edge == -1 ? dg_.N() : to_right[front_edge]);
    if (front_edge == -1) dg_.AddVertices(1);
    if (alternative_path_and_start.solo()) {
        dg_.AddEdge(first_node, last_node, EdgeLoc(alternative_path_and_start[0]));
    } else {
        int start_new_vertices = dg_.N();
        dg_.AddVertices(alternative_path_and_start.size() - 1);
        dg_.AddEdge(first_node, start_new_vertices, 
                EdgeLoc(alternative_path_and_start.front()));
        for (size_t i = 1; i < alternative_path_and_start.size() - 1; ++i) {
            dg_.AddEdge(start_new_vertices + i-1, start_new_vertices+i, 
                    EdgeLoc(alternative_path_and_start[i]));
        }
        dg_.AddEdge(dg_.N()-1, last_node, EdgeLoc(alternative_path_and_start.back()));
    }
}

// Align each edge from bubble graph to reference and call variants.
// Also report the alignment of these edges, which are used to find friend
// locations.
void EdgesOnRef::CallVariantsGroupedWithProb(int gid, vec<VariantCallGroup> *p_groups, 
        vec<align>* p_edge_aligns, int verbosity) 
{
    if (verbosity >=1) cout << Date() << ": Start calling variants" << endl;
    // identify anchoring edges from bubble graph
    vec<int> anchoring_edges;
    vec<pair<bool,bool>> free_ends;
    for (int n1 = 0; n1 < bubble_graph_.N(); n1++) {
        if (bubble_graph_.From(n1).size() == 1) {
            int n2 = bubble_graph_.From(n1)[0];
            int e = bubble_graph_.EdgeObjectIndexByIndexFrom(n1,0);
            ForceAssertEq(bubble_graph_.To(n2).size(), 1u);
            anchoring_edges.push_back(e);
            free_ends.push(bubble_graph_.To(n1).empty(),
                    bubble_graph_.From(n2).empty());
        }
    }
    int nedges = bubble_graph_.EdgeObjectCount();
    int ngroups = anchoring_edges.size() * 2 - 1;
    ForceAssert( ngroups >= 0 );
    // if ( ngroups < 0 ) ngroups = 0;

    // to be calculated 
    vec<VariantCallGroup> all_groups(ngroups);
    vec<align>   align_all_edges(nedges);
    for (size_t i = 0; i < anchoring_edges.size(); i++) {
        int e = anchoring_edges[i];
        const vec<int>& path = bubble_graph_edge_path_[e];
        vec<int> path0 = ConvertToAssemblyPathFromUnrolled(path);
        basevector edge_seq = bubble_graph_.EdgeObject(e);
        // trim K-1 base if overlaping with cells
        int trim_start = (free_ends[i].first ? 0 : hb_.K()-1);
        int trim_end = (free_ends[i].second ? 0 : hb_.K()-1); 

        char edge_prev_base = ( trim_start == 0 ? 'X' : 
                as_base(edge_seq[trim_start-1]) );
        edge_seq.SetToSubOf(edge_seq, trim_start, 
                edge_seq.size() - trim_start - trim_end);
        int path_offset = GetStart(path.front()) + trim_start;

        align a; int nerror = 0;
        int bandwidth = max(1, (int)edge_seq.size() /2 );
        SmithWatAffineBanded(edge_seq, gplus_, -path_offset, bandwidth, a, nerror);
        if (verbosity >= 2) {
            cout << "anchoring edge "  << i << " [" <<  a.pos1() << "," 
                << a.Pos1() << ") aligns to [" << a.pos2() << "," 
                << a.Pos2() << ") " << endl;
            cout << "path0= ";
            path0.Println(cout);
            if (trim_start == 0 || trim_end == 0) 
                cout << "trims= " << trim_start << "," << trim_end << endl;
            PrintVisualAlignment(True, cout, edge_seq, gplus_, a);
        }
        align_all_edges[e] = a;
        align_all_edges[e].Setpos2(a.pos2() - gplus_ext_); // related to G

        // find the edits
        vec<triple<int,int,String>> edits;
        vec<pair<String,String>> change;
        bool head_ins = (trim_start != 0);
        bool tail_ins = (trim_end != 0);
        GetEditsFromAlign(edge_seq, gplus_, a, &edits, &change, 
                edge_prev_base, 'X', head_ins, tail_ins);
        FilterAndModifyEdits(edits, change);
        vec<VariantCall> vcalls;
        for (size_t j = 0; j < edits.size(); j++) {
            vec<pair<int,int>> edge0_pos = FindEdgeHome(edits[j].first+trim_start, path0);
            for (size_t kk = 0; kk < edge0_pos.size(); kk++) {
                VariantCall vcall = { { gid, edits[j].second, edits[j].third, 
                                        change[j].first, change[j].second}, 
                                      { e, edits[j].first, 
                                        edge0_pos[kk].first, edge0_pos[kk].second} }; 
                vcalls.push_back(vcall);
            }
        }
        VariantCallGroup vcall_group(gid, a.pos2(), a.Pos2());
        a.Setpos2(a.pos2() - gplus_ext_); // related to G
        //new_hb_edge_mapping_.ConvertPathToOld(path0);
        vcall_group.AddBranch(path0, vcalls, a, trim_start, bubble_graph_edge_weight_[e]);
        vcall_group.AddRefWeight(vec<double>(bubble_graph_edge_weight_[e].size(), 0.0));
        all_groups[i*2] = vcall_group;
    }

    // find all paths between cell region
    int ncells = anchoring_edges.size() - 1;

    vec<int> to_right3; bubble_graph_.ToRight(to_right3);
    vec<int> to_left3;  bubble_graph_.ToLeft(to_left3);
    for ( int64_t i = 0; i < anchoring_edges.jsize()-1; i++ ) {
        int header_edge = anchoring_edges[i];
        int trailer_edge =  anchoring_edges[i+1];
        const basevector& header = bubble_graph_.EdgeObject(header_edge);
        int n1 = to_right3[header_edge];
        int n2 = to_left3[trailer_edge];
        int nbranches = bubble_graph_.From(n1).size();
        ForceAssertEq(bubble_graph_.From(n1).size(), bubble_graph_.To(n2).size());
        int cell_start = all_groups[2*i].GetEndPos();
        int cell_end = all_groups[2*i+2].GetStartPos();
        if (cell_start >= cell_end) {
            cout << "   skip the overlapping region between "
                << header_edge << " and " << trailer_edge << endl;
            continue;
        }
        basevector cell_ref(gplus_, cell_start, cell_end - cell_start);

        VariantCallGroup vcall_group(gid, cell_start, cell_end);
        vec<double> ref_weight;
        for (int j = 0; j < nbranches; j++) {
            int e = bubble_graph_.EdgeObjectIndexByIndexFrom(n1, j);
            vec<int> path0 =
                ConvertToAssemblyPathFromUnrolled(bubble_graph_edge_path_[e]);
            basevector branch_base = bubble_graph_.EdgeObject(e);

            // Align the edge to reference. penalize end-gaps during alignment
            int nerror; alignment al;
            int score = SmithWatAffine(branch_base, cell_ref, al, true, true);
            align a = align(al);
            if (verbosity >= 2) {
                cout << "align branch"  << j << " [" <<  a.pos1() << "," << a.Pos1() 
                    << ") aligns to [" << a.pos2() << "," << a.Pos2() << ") " << endl;
                cout << "path0= ";
                path0.Println(cout);
                PrintVisualAlignment(True, cout, branch_base, cell_ref, a);
            }
            a.Setpos2(a.pos2() + cell_start);
            align_all_edges[e] = a;
            align_all_edges[e].Setpos2(a.pos2() - gplus_ext_); // related to G

            vec<triple<int,int,String>> edits;
            vec<pair<String,String>> change;
            char prev_base = as_base(*(header.end() - bubble_graph_.K()));
            GetEditsFromAlign(branch_base, gplus_, a, &edits, &change, 
                    prev_base, 'X', true, true);
            FilterAndModifyEdits(edits, change);

            vec<VariantCall> vcalls;
            for (size_t k = 0; k < edits.size(); k++) {
                vec<pair<int,int>> edge0_pos = FindEdgeHome(edits[k].first, path0);
                for (size_t kk = 0; kk < edge0_pos.size(); kk++) {
                    VariantCall vcall = { { gid, edits[k].second, edits[k].third, 
                                            change[k].first, change[k].second }, 
                                          { e, edits[k].first, 
                                            edge0_pos[kk].first, edge0_pos[kk].second } }; 
                    vcalls.push_back(vcall);
                }
            }
            if (bubble_graph_edge_is_assembly_[e]) {
                a.Setpos2(a.pos2() - gplus_ext_); // related to G
                //new_hb_edge_mapping_.ConvertPathToOld(path0);
                vcall_group.AddBranch(path0, vcalls, a, 0, bubble_graph_edge_weight_[e]);
            }
            if (bubble_graph_edge_is_ref_[e]) 
                ref_weight = bubble_graph_edge_weight_[e];
        }
        vcall_group.AddRefWeight(ref_weight);
        all_groups[2*i+1] = vcall_group;
    }
    copy(all_groups.begin(), all_groups.end(), back_inserter(*p_groups));
    copy(align_all_edges.begin(), align_all_edges.end(), back_inserter(*p_edge_aligns));

    if (verbosity >=1) cout << Date() << ": Ended calling variants" << endl;
};

// Determine the location on sub-edges given a position in the untrimmed path edge,
// The sub-edges are the edge id from hbp
vec<pair<int,int>> EdgesOnRef::FindEdgeHome(int pos_on_edge, const vec<int>& path0) const
{
    vec<pair<int,int>> edge0_start;
    int sub_start = 0, sub_end = 0;
    for (size_t i = 0; i < path0.size(); i++) {
        sub_start = (i == 0 ? 0 : sub_end - hb_.K() + 1);
        sub_end = sub_start + hb_.EdgeLengthBases(path0[i]);
        if (sub_start >= pos_on_edge) break;
        if (pos_on_edge < sub_end ) 
            edge0_start.push(path0[i], pos_on_edge - sub_start);
    }
    for (auto& p: edge0_start) {
        new_hb_edge_mapping_.ConvertEdgeIdAndOffsetToOld(p);
    }
    auto end = std::unique(edge0_start.begin(), edge0_start.end());
    edge0_start.resize(distance(edge0_start.begin(), end));
    return edge0_start;
}

void EdgesOnRef::DumpUnrolled(String filename, 
        const vec<pair<int,Bool>>* p_hbp_to_hb) const
{
    vec<String> edge_names2(dg_.EdgeObjectCount(),"");
    vec<double> lengths2( dg_.EdgeObjectCount( ) );
    for (size_t i = 0; i < edge_names2.size(); ++i) {
        int eid = GetOriginalId(i);
        if (p_hbp_to_hb != NULL)
            eid = (*p_hbp_to_hb)[eid].first;
        edge_names2[i] = ToString(i) + "->" + ToString(eid) + "@"
            + ToString(GetStart(i));
        lengths2[i] = hb_.EdgeLengthKmers(GetOriginalId(i));
    }
    ofstream dout(filename);
    dg_.PrettyDOT( dout, lengths2, digraphE<EdgeLoc>::edge_label_info(digraphE<EdgeLoc>
                ::edge_label_info::DIRECT, &edge_names2 ) );
}

// Linearize the graph to a string of bubbles connected by single-edge 
// anchoring edges. 
void EdgesOnRef::MakeBubbleGraph(int verbosity) 
{
    if (verbosity >=1)
        cout << Date() << ": Make bubble graph" << endl;
    vec<vec<int>> single_edges;
    vec<align> single_edges_aligns;
    FindAnchoringEdges(single_edges);
    AlignAnchoringEdges(single_edges, single_edges_aligns, verbosity);
    if (verbosity >=2)
        cout << "build the graph backbone" << endl;
    bubble_graph_.Clear();
    bubble_graph_.SetK(hb_.K());
    bubble_graph_.AddVertices(single_edges.size() * 2);
    for (size_t i = 0; i < single_edges.size(); i++) {
        vec<int> path0 = ConvertToAssemblyPathFromUnrolled(single_edges[i]);
        basevector edge = hb_.EdgePathToBases(path0);
        bubble_graph_.AddEdge(i*2, i*2+1, edge);
        bubble_graph_edge_path_.push_back(single_edges[i]);
        bubble_graph_edge_is_ref_.push_back(false);
        bubble_graph_edge_is_assembly_.push_back(true);
    }
    if (verbosity >=2)
        cout << "build the graph bubbles" << endl;
    for ( int64_t i = 0; i < single_edges.jsize()-1; i++ ) {
        int ref_start = single_edges_aligns[i].Pos2();
        int ref_stop = single_edges_aligns[i+1].pos2();

        int entrance_edge = single_edges[i].back();
        int exit_edge     = single_edges[i+1].front();
        const int NMaxPath = 50;
        vec<vec<int>> cell_paths;
        FindAllPathsNoLoop(dg_, entrance_edge, exit_edge, &cell_paths, NMaxPath);
        if (cell_paths.empty()) continue;
        if (ref_stop < ref_start) continue;

        basevector ref_seg;
        ref_seg.SetToSubOf(gplus_, ref_start, ref_stop-ref_start);

        bool has_ref = false;
        for (size_t j = 0; j < cell_paths.size(); j++) {
            vec<int> path0 = ConvertToAssemblyPathFromUnrolled(cell_paths[j]);
            basevector edge = hb_.EdgePathToBases(path0);
            bubble_graph_.AddEdge(i*2+1, i*2+2, edge);
            bubble_graph_edge_path_.push_back(cell_paths[j]);
            bubble_graph_edge_is_ref_.push_back(edge == ref_seg);
            bubble_graph_edge_is_assembly_.push_back(true);
            if (edge == ref_seg) { has_ref = true; }
        }
        if (!has_ref) {
            bubble_graph_.AddEdge(i*2+1, i*2+2, ref_seg);
            bubble_graph_edge_path_.push_back(vec<int>());
            bubble_graph_edge_is_ref_.push_back(true);
            bubble_graph_edge_is_assembly_.push_back(false);
        }
    }
    ForceAssertEq(bubble_graph_edge_path_.isize(), bubble_graph_.EdgeObjectCount());
    bubble_graph_edge_weight_.assign(bubble_graph_edge_path_.size(), vec<double>());
}

void EdgesOnRef::DumpBubbleGraph(String filename) const
{
    const vec<vec<int>>& supports = bubble_graph_edge_supports_;
    vec<double> lengths(bubble_graph_.EdgeObjectCount());
    vec<String> edge_names(bubble_graph_.EdgeObjectCount());
    for (size_t e = 0; e < lengths.size(); e++) {
        lengths[e] = bubble_graph_.EdgeLengthKmers(e);
        edge_names[e] += ToString(e) + ": ";
        for (size_t k = 0; k < supports[e].size(); k++) {
            if (k != 0 )
                edge_names[e] += ",";
            edge_names[e] += ToString(supports[e][k]);
        }
    }
    ofstream dout(filename);
    bubble_graph_.PrettyDOT(dout, lengths, HyperBasevector::edge_label_info(
                HyperBasevector::edge_label_info::DIRECT, &edge_names) );
}

// Calculate probability of being real for each paths.
void EdgesOnRef::PathProb(const vecbasevector& bases, const vecqualvector& quals, 
        int verbosity) 
{
    if (verbosity >= 1) 
        cout << Date() << ": Calculate probabilities of edges" << endl;
    
    // { { (rid11, diff11), (rid12, diff12), ...} // for edge1
    //   { (rid21, diff21), (rid22, diff22), ...} // for edge2
    //   ...  // }
    vec<vec<pair<int,int>>> homes_index(bubble_graph_.EdgeObjectCount());
    FindReadHomesBest(bases, quals, bubble_graph_, &homes_index, 
            vec<int>(), NULL, verbosity);
    vec<String> samples = p_read_tracker_->getSampleList();
    int nsamples = p_read_tracker_->getSampleList().size();
    if (verbosity >= 2) {
        for (size_t i = 0; i < samples.size(); i++) 
            cout << "sample " << i << " " << samples[i] << endl;
    }
    vec<vec<int>> supports(bubble_graph_.EdgeObjectCount(), vec<int>(nsamples,0));
    vec<vec<double>> probs(bubble_graph_.EdgeObjectCount(), vec<double>(nsamples,0.0));
    for (int n1 = 0; n1 < bubble_graph_.N(); n1++) {
        if (bubble_graph_.From(n1).size() < 1) continue;
        int n2 = bubble_graph_.From(n1)[0];
        vec<int> branches(bubble_graph_.From(n1).size());
        for (size_t j = 0; j < bubble_graph_.From(n1).size(); j++) 
            branches[j] = bubble_graph_.EdgeObjectIndexByIndexFrom(n1,j);
        const int qgood = 4;
        for (size_t j = 0; j < branches.size(); j++) {
            int e = branches[j];
            if (verbosity>=2) {
                cout << "branches " << j << " edge= " << e << endl;
                for (auto& x: homes_index[e]) {
                    int rid = x.first;
                    int sample_id = p_read_tracker_->getSampleID(rid);
                    if (x.second/1000 >= qgood && sample_id >= 0) 
                        cout << x.first << "," << x.second << " ";
                }
                cout << endl;
            }
            for (auto& x: homes_index[e]) {
                int rid = x.first;
                int sample_id = p_read_tracker_->getSampleID(rid);
                if (x.second/1000 >= qgood && sample_id >= 0) {
                    supports[e][sample_id] += x.second/1000;    
                    probs[e][sample_id] += x.second/1000;    
                }
            }
        }
    }
    bubble_graph_edge_weight_ = probs;
    bubble_graph_edge_supports_ = supports;
    if (verbosity >=1 ) 
        DumpBubbleGraph("bubble.dot");
}

// Find all paths between node n1 and n2 from graph dg_. Do not allow
// traversal of same edges twice

// auxilary data structure for FindAllPathsNoLoop
namespace {
struct PartialPath {
    vec<int> nodes;
    vec<int> edges;
};
}

template<typename GraphT>
void EdgesOnRef::FindAllPathsNoLoop(const GraphT& dg, int entrace_edge, int exit_edge, 
        vec<vec<int>>* allpaths, const int NMaxPath) const
{
    const int MaxDepth = 100;
    vec<int> to_right2; dg.ToRight(to_right2);
    vec<int> to_left2; dg.ToLeft(to_left2);
    int n1 = to_right2[entrace_edge];
    int n2 = to_left2[exit_edge];

    PartialPath start = {{n1},{}};
    stack<PartialPath> visited;
    visited.push(start);
    while (! visited.empty()) {
        if ((int)allpaths->size() > NMaxPath) break;
        PartialPath p0 = visited.top();
        visited.pop();
        int node = p0.nodes.back();
        if (node == n2) { allpaths->push_back(p0.edges); }
        if ((int)p0.edges.size() < MaxDepth) {
            for (int i = dg.From(node).size() -1; i >= 0; i--){
                int node_new = dg.From(node)[i];
                int edge_new = dg.EdgeObjectIndexByIndexFrom(node, i);
                if (!Member(p0.edges, edge_new)
                        && edge_new != exit_edge) {
                    PartialPath pnew = p0;
                    pnew.nodes.push_back(node_new);
                    pnew.edges.push_back(edge_new);
                    visited.push(pnew);
                }
            }
        }
    }
}

// Delete edits that are outof range. Also make corrections to 
// position of edits as original alignment was with Gplus.
// Add change the position of insertion and deletion to previous base.
void EdgesOnRef::FilterAndModifyEdits( vec<triple<int,int,String>>& edits, 
        vec<pair<String,String>>& change)
{
    ForceAssertEq(edits.size(), change.size());
    // only take edits in the genome_ region
    vec<Bool> todel(edits.size(), false);
    for (size_t j = 0; j < edits.size(); j++) {
        edits[j].second -= gplus_ext_;
        if (change[j].first.size() != change[j].second.size()) {
            edits[j].second -= 1;
            edits[j].first -= 1;
        }
        if ( edits[j].second < 0 || edits[j].second >= (int)genome_.size())
            todel[j] = True; 
    }
    EraseIf(edits, todel);
    EraseIf(change, todel);
}

vec<int> EdgesOnRef::ConvertToAssemblyPathFromUnrolled(const vec<int>& unrolled_path) const
{
    vec<int> path0;
    for (auto& x: unrolled_path) path0.push_back(GetOriginalId(x)); 
    return path0;
}

// Prepair backbone of the bubble graph. The backbone is single edged part in
// the graph.  Also remove any edge if the length is shorter than K after
// trimming the overlapping K-1 bases on both ends, or cannot be anchored in
// the genome (to be implemented). 
void EdgesOnRef::FindAnchoringEdges(vec<vec<int>>& single_edges) const
{
    single_edges.clear();
    vec<int> to_right2; dg_.ToRight(to_right2);
    vec<int> to_left2; dg_.ToLeft(to_left2);
    // backboen that are not covered by alternative path
    vec<int> backbone;
    for (size_t i = 0; i < backbone_.size(); i++) 
        if (backbone_dup_[i] == 0) backbone.push_back(backbone_[i]);
    // stitch continuous backbone edges
    for (size_t i = 0; i < backbone.size(); i++) {
        vec<int> path = {backbone[i]};
        size_t j = i + 1; // extension
        while (j < backbone.size()) {
            int node1 = to_right2[backbone[j-1]];
            int node2 = to_left2[backbone[j]];
            if (node1 != node2) {
                vec<vec<int>> allpaths;
                FindAllPathsNoLoop(dg_, backbone[j-1], backbone[j], &allpaths, 3);
                if (allpaths.size() != 1) break;
                else path.append(allpaths[0]);
            }
            path.push_back(backbone[j]);
            j++;
        }
        single_edges.push_back(path);
        i = j-1;
    }
}

// Remove short edges. Also remove edges that cannot be anchored in reference,
// which are probably due to large insertions.
void EdgesOnRef::AlignAnchoringEdges(vec<vec<int>>& single_edges,
        vec<align>& single_edge_aligns, int verbosity) 
{
    vec<int> to_right2; dg_.ToRight(to_right2);
    vec<int> to_left2; dg_.ToLeft(to_left2);
    single_edge_aligns.clear();
    single_edge_aligns.resize(single_edges.size());
    vec<Bool> todel(single_edges.size(), False);
    const int MinAlignedBase = 20;
    for (size_t i = 0; i < single_edges.size(); i++) {
        const vec<int>& path = single_edges[i];    // path in dg_
        vec<int> path0 = ConvertToAssemblyPathFromUnrolled(path);
        basevector edge = hb_.EdgePathToBases(path0);
        int trim_left = 0, trim_right = 0;
        if (dg_.To(to_left2[single_edges[i].front()]).size() != 0)
            trim_left = hb_.K() - 1;
        if (dg_.From(to_right2[single_edges[i].back()]).size() != 0)
            trim_right = hb_.K() - 1;
        if ((int)edge.size() - trim_left - trim_right < MinAlignedBase) {
            todel[i] = True;
            if (verbosity >= 2) {
                cout << "Erasing short anchoring edge " << i << " len=" 
                    << edge.size() << " trim_left= " << trim_left 
                    << " trim_right= " << trim_right << endl;
                cout << " path= ";
                path.Println(cout);
                cout << " path0= ";
                path0.Println(cout);
            }
        } else {
            // make the alignment
            basevector edge_trimmed(edge.begin()+trim_left, 
                    edge.end() - trim_right);
            align a; int nerror = 0;
            int bandwidth = max(1, (int)edge_trimmed.size() /2 );
            int path_offset = GetStart(path.front()) + trim_left;
            SmithWatAffineBanded(edge_trimmed, gplus_, -path_offset, bandwidth, a, nerror);
            single_edge_aligns[i] = a;
            if (verbosity >=2) {
                cout << "anchoring edge " << i << " trim_left= " << trim_left
                    << " trim_right= " << trim_right  << " path0= ";
                path0.Println(cout);
                cout << "    found alignment " << " [" <<  a.pos1() << "," << a.Pos1() << ") to [" 
                    << a.pos2() << "," << a.Pos2() << ") " << endl;
            }
            if (a.Pos1() - a.pos1() < MinAlignedBase ||
                    a.Pos2() - a.pos2() < MinAlignedBase ){
                todel[i] = True;
                if (verbosity >=2) {
                    cout << "    Unable to align anchoring edge " << i << "to reference: "
                        << " alignment= " << " [" <<  a.pos1() << "," << a.Pos1() << ") to [" 
                        << a.pos2() << "," << a.Pos2() << ") " << endl;
                    cout << "   path0= ";
                    path0.Println(cout);
                    cout << "   path= ";
                    path.Println(cout);
                }
            } 
        }
    }
    if (todel.CountValue(True) > 0) {
        if (verbosity >= 2) {
            cout << "Erasing short anchoring edges:  " << endl;
            for (size_t i = 0; i < todel.size(); i++) {
                if (todel[i]) {
                    cout << " i= " << i << " path= ";
                    single_edges[i].Println(cout);
                }
            }
        }
        EraseIf(single_edges, todel);
        EraseIf(single_edge_aligns, todel);
    }
    ForceAssertEq(single_edges.size(), single_edge_aligns.size());
}

// explicit instantiation of some digraph functions used locally
template void digraphE<EdgeLoc>::AddEdge( const int v, const int w, const EdgeLoc& e );
template void digraphE<EdgeLoc>::AddVertices(int nadd);
template void digraphE<EdgeLoc>::ToLeft(vec<int>& to_left) const;
template void digraphE<EdgeLoc>::ToRight(vec<int>& to_right) const;
